%{

/*
   cfg.l -- PIKT piktc lexer for all .cfg files except systems.cfg

   $Header: /home/cvs/ucgsb/pikt/src/piktc/cfg.l,v 1.8 1999/04/15 04:23:39 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "common.h"
#include "cfg.tab.h"

int ii;
int fn = 0;

int context;	/* STRING/NUMBER */

int cfginput(void);
int condparen(char *);
int fnstr(char *);

#define RETURN(token) cfglval.sval = strdup(yytext); return token

#undef YY_INPUT
#define YY_INPUT(buf, result, max_size) { \
		char c = cfginput(); \
		result = (c == EOF) ? YY_NULL : (buf[0] = c, 1); \
	}

%}

W		[ \t]
D		[0-9]
L		[A-Za-z]
DL		[0-9A-Za-z]

%x MS AS PS OS CS DS SS FS

%option noyywrap

%%

"/"{3,7}\n			{ ii = 0;
				  switch(strlen(yytext)-1) {
					case 3:
						return _MCRCFGBOF;
						break;
					case 4:
						return _ALMCFGBOF;
						break;
					case 5:
						return _ALTCFGBOF;
						break;
					case 6:
						return _PRGCFGBOF;
						break;
					case 7:
						return _OBJCFGBOF;
						break;
				  }
				}

<MS,AS,PS,OS>"/"{3,7}\n		{ yyless(0);
				  BEGIN(INITIAL);
				}

<MS,AS,PS,OS>#"/"{3,7}\n	{ yyless(1);
				  BEGIN(INITIAL);
				}

<*>^{D}+{W}*\n			{ lineno = atoi(yytext); }

<*>^#				{ ; /* dummy char; see
				       cfginput() below
				     */
				}

<*>^({L}|"/")[-_\./0-9A-Za-z]*	{ if ((yytext[0] == '/') && (cfgcurr != PRGCFG))
					cfgerror("syntax error");
				  switch(cfgcurr) {
				  	case MCRCFG:
						BEGIN(MS);
						break;
					case ALTCFG:
						context = NIL;
						BEGIN(AS);
						break;
					case PRGCFG:
						BEGIN(PS);
						break;
					case OBJCFG:
						BEGIN(OS);
						break;
					default:
					  	BEGIN(INITIAL);
						break;
				  }
				  RETURN(_STANZA);
				}

<INITIAL,DS,SS>status		{ RETURN(_STATUS); }
<INITIAL,DS,SS>active		{ RETURN(_ACTIVE); }
<INITIAL,DS,SS>inactive		{ RETURN(_INACTIVE); }

<INITIAL,DS,SS>level		{ RETURN(_LEVEL); }
<INITIAL,DS,SS>emerg|emergency	{ RETURN(_EMERG); }
<INITIAL,DS,SS>alert|urgent	{ RETURN(_ALERT); }
<INITIAL,DS,SS>crit|critical	{ RETURN(_CRIT); }
<INITIAL,DS,SS>err|error	{ RETURN(_ERR); }
<INITIAL,DS,SS>warn|warning	{ RETURN(_WARNING); }
<INITIAL,DS,SS>notice		{ RETURN(_NOTICE); }
<INITIAL,DS,SS>info		{ RETURN(_INFO); }
<INITIAL,DS,SS>debug		{ RETURN(_DEBUG); }

<INITIAL,DS,SS>task		{ RETURN(_TASK); }

<INITIAL,DS,SS>input		{ RETURN(_INPUT); }
<INITIAL,DS,SS>proc|process	{ BEGIN(CS);
				  RETURN(_PROCESS);
				}
<INITIAL,DS,SS>file		{ RETURN(_REGFILE); }
<INITIAL,DS,SS>logfile		{ RETURN(_LOGFILE); }

<INITIAL,DS,SS>filter		{ BEGIN(CS);
				  RETURN(_FILTER);
				}

<INITIAL,DS,SS>seps|separators	{ RETURN(_SEPARATORS); }
<INITIAL,DS,SS>dat|data		{ BEGIN(DS);
				  RETURN(_DAT);
				}
<INITIAL,DS,SS>keys		{ BEGIN(DS);
				  RETURN(_KEYS);
				}

<INITIAL,DS,SS>init		{ RETURN(_INIT); }
<INITIAL,DS,SS,CS>begin		{ BEGIN(INITIAL);
				  RETURN(_BEGIN);
				}
<INITIAL,DS,SS,CS>end		{ BEGIN(INITIAL);
				  RETURN(_END);
				}
<INITIAL,DS,SS,CS>rule		{ BEGIN(INITIAL);
				  RETURN(_RULE);
				}

<INITIAL,DS,SS>do		{ context = NIL;
				  RETURN(_DO);
				}
<INITIAL,DS,SS>if		{ context = NUMBER;
				  RETURN(_IF);
				}
<INITIAL,DS,SS>if{W}*"("	{ context = NUMBER;
				  unput('[');
				  cfglval.sval = strdup("if");
				  return _IF;
				}
<INITIAL,DS,SS>(elif|elsif|elseif) {
				  context = NUMBER;
				  RETURN(_ELIF);
				}
<INITIAL,DS,SS>(elif|elsif|elseif){W}*"(" {
				  context = NUMBER;
				  unput('[');
				  cfglval.sval = strdup("elif");
				  return _ELIF;
				}
<INITIAL,DS,SS>else		{ context = NIL;
				  RETURN(_ELSE);
				}
<INITIAL,DS,SS>endif|fi		{ context = NIL;
				  RETURN(_ENDIF);
				}
<INITIAL,DS,SS>while		{ context = NUMBER;
				  RETURN(_WHILE);
				}
<INITIAL,DS,SS>while{W}*"("	{ context = NUMBER;
				  unput('[');
				  cfglval.sval = strdup("while");
				  return _WHILE;
				}
<INITIAL,DS,SS>endwhile		{ context = NIL;
				  RETURN(_ENDWHILE); }
<INITIAL,DS,SS>repeat		{ context = NIL;
				  RETURN(_REPEAT);
				}
<INITIAL,DS,SS>until		{ context = NUMBER;
				  RETURN(_UNTIL); }
<INITIAL,DS,SS>until{W}*"("	{ context = NUMBER;
				  unput('[');
				  cfglval.sval = strdup("until");
				  return _UNTIL;
				}
<INITIAL,DS,SS>for		{ context = NUMBER;
				  RETURN(_FOR);
				}
<INITIAL,DS,SS>for{W}*"("	{ context = NUMBER;
				  RETURN(_FORLP);
				}
<INITIAL,DS,SS>endfor		{ context = NIL;
				  RETURN(_ENDFOR);
				}
<INITIAL,DS,SS>break		{ context = NIL;
				  RETURN(_BREAK);
				}
<INITIAL,DS,SS>continue|cont	{ context = NIL;
				  RETURN(_CONTINUE);
				}
<INITIAL,DS,SS>switch		{ context = NIL;
				  RETURN(_SWITCH);
				}
<INITIAL,DS,SS>switch{W}*"("	{ context = NIL;
				  RETURN(_SWITCHLP);
				}
<INITIAL,DS,SS>case		{ context = NIL;
				  RETURN(_CASE);
				}
<INITIAL,DS,SS>default		{ context = NIL;
				  RETURN(_DEFAULT);
				}
<INITIAL,DS,SS>breakswitch|breaksw {
				  context = NIL;
				  RETURN(_BREAKSWITCH);
				}
<INITIAL,DS,SS>endswitch|endsw	{ context = NIL;
				  RETURN(_ENDSWITCH);
				}
<INITIAL,DS,SS>again		{ context = NIL;
				  RETURN(_AGAIN);
				}
<INITIAL,DS,SS>redo		{ context = NIL;
				  RETURN(_REDO);
				}
<INITIAL,DS,SS>next		{ context = NIL;
				  RETURN(_NEXT);
				}
<INITIAL,DS,SS>skip		{ context = NUMBER;
				  RETURN(_SKIP);
				}
<INITIAL,DS,SS>last		{ context = NIL;
				  RETURN(_LAST);
				}
<INITIAL,DS,SS>pause		{ context = NUMBER;
				  RETURN(_PAUSE);
				}
<INITIAL,DS,SS>quit		{ context = NIL;
				  RETURN(_QUIT);
				}
<INITIAL,DS,SS>die		{ context = STRING;
				  RETURN(_DIE);
				}

<INITIAL,DS,SS>output		{ context = STRING;
				  RETURN(_OUTPUT);
				}
<INITIAL,DS,SS>mail		{ RETURN(_MAIL); }
<INITIAL,DS,SS>print		{ RETURN(_PRINT); }
<INITIAL,DS,SS>log		{ RETURN(_LOG); }
<INITIAL,DS,SS>piktlog		{ RETURN(_PIKTLOG); }
<INITIAL,DS,SS>syslog		{ RETURN(_SYSLOG); }

<INITIAL,DS,SS>set		{ context = NIL;
				  BEGIN(SS);
				  RETURN(_SET);
				}

<INITIAL,DS,SS>exec		{ context = STRING;
				  BEGIN(INITIAL);
				  RETURN(_EXEC);
				}
<INITIAL>wait			{ RETURN(_WAIT); }

<CS,DS>\"[^\n]+\"/[\n\t ]*	{ BEGIN(INITIAL);
				  RETURN(_STR);
				}

<DS>({D}{1,5}|$|$-{D}{1,5})/[^\]a-zA-Z0-9] {
				  BEGIN(INITIAL);
				  RETURN(_POSITION);
				}
<DS>"["({D}{1,5}|$|$-{D}{1,5})"]" { BEGIN(INITIAL);
				  RETURN(_POSITION);
				}
<DS>{D}+","({D}+|$)		{ BEGIN(INITIAL);
				  RETURN(_RANGE);
				}
<DS>"["{D}+","({D}+|$)"]"	{ BEGIN(INITIAL);
				  RETURN(_RANGE);
				}

<INITIAL,SS>"="			{ context = NIL;
				  BEGIN(INITIAL);
				  return yytext[0];
				}
<INITIAL,SS>("+="|"-="|"*="|"/="|"%="|"**="|"<<="|">>="|"&="|"|="|"^=") {
				  context = NUMBER;
				  BEGIN(INITIAL);
				  RETURN(_AOPNUM);
				}
<INITIAL,SS>".="		{ context = STRING;
				  BEGIN(INITIAL);
				  RETURN(_AOPSTR);
				}

<INITIAL>\"[^\"]*\"		{ RETURN(_STR);	}

<FS>[A-Z][A-Z0-9]*		{ RETURN(_FLHNDL); }

<DS,SS,FS,INITIAL>(#|$){L}{DL}*	{ context = (yytext[0] == '#' ? NUMBER : STRING);
				  RETURN(yytext[0] == '#' ? _VARNUM : _VARSTR);
				}

<FS,INITIAL>(%|@){L}{DL}*/("["[^\n]+"]")*{W}*("+"|"-"|"*"|"/"|"%"|"**"|"& "|"| "|"^"|"<<"|">>"|"=="|">"|"<"|">="|"<="|"!>"|"!<"|"!="|"<>"|"]"){W}* {
				  context = NUMBER;
				  RETURN(yytext[0] == '%' ? _VARHSTNUM : _VARPRCNUM);
				}

<FS,INITIAL>(%|@){L}{DL}*/("["[^\[\]]+"]")*{W}*("."|"=~"|"=~~"|"!~"|"!~~"){W}* {
				  context = STRING;
				  RETURN(yytext[0] == '%' ? _VARHSTSTR : _VARPRCSTR);
				}

<FS,INITIAL>(%|@){L}{DL}*/("["[^\[\]]+"]")*{W}+(eq|ne|lt|gt|le|ge){W}* {
				  context = STRING;
				  RETURN(yytext[0] == '%' ? _VARHSTSTR : _VARPRCSTR);
				}

<FS,INITIAL>(%|@){L}{DL}*	{ switch (context) {
					case NUMBER:
						RETURN(yytext[0] == '%' ? _VARHSTNUM : _VARPRCNUM);
						break;
					case STRING:
					case NIL:
					default:
						RETURN(yytext[0] == '%' ? _VARHSTSTR : _VARPRCSTR);
						break;
				  }
				}

<SS,FS,INITIAL>${D}{1,5}	{ context = STRING;
				  RETURN(_VALSTR);
				}

<SS,FS,INITIAL>$"["		{ yytext[yyleng-1] = '\0';
				  cfglval.sval = strdup(yytext);
				  unput('[');
				  unput('[');
				  unput('[');
				  context = STRING;
				  return _VALSTR;
				}

<DS>(#|$){L}{DL}*"["		{ yytext[yyleng-1] = '\0';
				  cfglval.sval = strdup(yytext);
				  unput('[');
				  unput('[');
				  unput('[');
				  context = (cfglval.sval[0] == '#' ? NUMBER : STRING);
				  return (cfglval.sval[0] == '#' ? _VARNUM : _VARSTR);
				}

<DS,SS,FS,INITIAL>"["		{ context = NUMBER;
				  return yytext[0];
				}
<DS,SS,FS,INITIAL>"[[["		{ return '['; }
<DS,SS,FS,INITIAL>{D}{1,5}	{ context = NUMBER;
				  RETURN(_INT);
				}
<DS,SS,FS,INITIAL>"]"		{ return yytext[0]; }

<SS,FS,INITIAL>("+"|"-"|"~")	{ context = NUMBER;
				  return yytext[0];
				}

<SS,FS,INITIAL>("*"|"/"|"%"|"**"|"&"|"|"|"^"|"<<"|">>") {
				  context = NUMBER;
				  RETURN(_NOP);
				}

<SS,FS,INITIAL>"."		{ context = STRING;
				  RETURN(_SOP);
				}

<SS,FS,INITIAL>(-e|-z|-f|-d|-c|-b|-p|-l|-S) {
				  context = STRING;
				  RETURN(_FOP);
				}

<SS,FS,INITIAL>("=="|">"|"<"|">="|"<="|"!>"|"!<"|"!="|"<>") {
				  context = NUMBER;
				  RETURN(_NREL);
				}
<SS,FS,INITIAL>("=~"|"=~~"|"!~"|"!~~"|eq|ne|lt|gt|le|ge) {
				  context = STRING;
				  RETURN(_SREL);
				}

<SS,FS,INITIAL>{D}{1,2}		{ context = NUMBER;
				  RETURN(_NUM);
				}
<SS,FS,INITIAL>(({D}+)|({D}*\.{D}+)|((({D}+)|({D}*\.{D}+))\/(({D}+)|({D}*\.{D}+))))%* {
				  context = NUMBER;
				  RETURN(_NUM);
				}

<SS,FS,INITIAL>(#|$){L}{DL}*/{W}*"(" {
				  context = (yytext[0] == '#' ? NUMBER : STRING);
				  BEGIN(FS);
				  RETURN(yytext[0] == '#' ? _FNNUM : _FNSTR);
				}
<SS,FS>"("			{ context = NIL;
				  fn++;
				  return '(';
				}
<SS,FS>")"			{ context = NIL;
				  fn--;
				  if (!fn)
					BEGIN(INITIAL);
				  return ')';
				}

<SS,FS>\"[^\n]*\"/[\n\t ]*	{ context = STRING;
				  ii = fnstr(yytext);
				  yyless(ii+1);
				  yytext[ii+1] = '\0';
				  RETURN(_STRFN);
				}

"("[\n\t ]*[^\n\t ]+[\n\t ]*(=~|=~~|!~|!~~|eq|ne|lt|gt|le|ge)[\n\t ]*[^\n\t ]+[\n\t ]*")"[\n\t ]* { 
				  context = STRING;
				  yyless(1);
				  return '[';
				}

"("([-+*/%$#@~^><)(.0-9a-zA-Z\n\t\[\] ]|"&"[^&]|"|"[^|])+ {
				  if (condparen(yytext)) {
					context = NUMBER;
					yyless(1);
					return '[';
				  } else {
					context = NUMBER;
					yyless(1);
					return yytext[0];
				  }
				}

")"				{ return yytext[0]; }

<SS,FS,INITIAL>"!"		{ context = NUMBER;
				  RETURN(_NOT);
				}
<SS,FS,INITIAL>"&&"		{ context = NUMBER;
				  RETURN(_AND);
				}
<SS,FS,INITIAL>"||"		{ context = NUMBER;
				  RETURN(_OR);
				}

<MS>[^ \t\n]+			{ RETURN(_MACRODEF); }
<MS>\"[^\"]*\"			{ RETURN(_MACRODEF); }

<AS>timing			{ return _TIMING; }
<AS>(priority|nice)		{ return _PRIORITY; }
<AS>nicecmd			{ return _NICECMD; }
<AS>mailcmd			{ return _MAILCMD; }
<AS>lpcmd			{ return _LPCMD; }
<AS>alarms			{ return _ALARMS; }
<AS>[-,*0-9]+			{ RETURN(_TIME); }
<AS>{D}{1,2}-{D}{1,2}		{ RETURN(_TIME); }
<AS>"*/"{D}{1,2}		{ RETURN(_TIME); }
<AS>{D}{1,2}-{D}{1,2}"/"{D}{1,2} { RETURN(_TIME); }
<AS>\"[^\"]*\"			{ yytext[yyleng-1] = '\0';
				  cfglval.sval = strdup(yytext+1);
				  return _STR;
				}
<AS>[^ \t\n]+			{ RETURN(_WORD); }

<PS>{W}/\n			{ cfglval.sval = strdup("");
				  return _LINE;
				}

<PS>{W}[^\n]+			{ cfglval.sval = strdup(yytext+1);
				  return _LINE;
				}

<OS>[^ \t\n][^\n]*		{ RETURN(_OBJECT); }

<*>","				{ return yytext[0]; }

<*>{W}				{ ; }
<*>\n				{ ; }

<*>.				{ return yytext[0]; }

%%

/* cfg error handler
 */
void
cfgerror (const char *message)
{

	if (stzerr)
		return;
	else {
		stzerr = TRUE;
		cfgerr = TRUE;
		if (cfp && strcmp(cfp,cfgbof[cfgcurr]) && (cfgcurr <= cfglast))
			fprintf(stderr, "%s, at or near line %d: %s (in cfgparse()): %s\n", cfgfilecurr, lineno, message, yytext);
		else
			fprintf(stderr, "%s, at or near line %d: %s (in cfgparse())\n", cfgfileprev ? cfgfileprev : cfgfilecurr, lineno, message);
		bailout = TRUE;
		BEGIN(INITIAL);
		return;
	}
}

/* for config files macros.cfg, alarms.cfg, alerts.cfg, programs.cfg &
   objects.cfg (not systems.cfg or defines.cfg), feed their contents to
   the lexer input stream; cfginput() is where we do macro substitutions
 */
int
cfginput (void)
{
	static char *cp;
	static cfgln *clp;
	static char linenum[8];
	static char *macroname;
	static ident *idp;
	static def *dfp;
	static char *mcp;
	static int escape;
	static int macro;
	static int macroffset;
	static int domcr;
	static unsigned offset_save;
	register int i;
	register char c;
	register char mc;

	if ((cfgcurr > cfglast) /* || (cfgcurr > OBJCFG) */) {
		stroffset = 0;
		return EOF;
	}

	c = ' ';	/* junk return value, just in case */
	if (!stroffset) {
		if (bof) {
			cfp = cp = cfgbof[cfgcurr];
			clp = configs[cfgcurr];
			cfgfileprev = cfgfilecurr;
			cfgfilecurr = cfgfiles[cfgcurr];
			/* zero this flag; at end of config lines, if this
			   is still zero, we know that no config lines were
			   processed for current host
			 */
			linenum[0] = '\0';
			escape = FALSE;
		} else if (bon) {
			while ((! bsmember(clp->bs, hbscurr, hbslongs)) || ! clp->doproc) {
				clp = clp->next;
				if (clp) {
					cfgfileprev = cfgfilecurr;
					cfgfilecurr = clp->file;
				} else {
					if (!linenum[0])
						logexit(PIKTLOG, ERR, __FILE__, __LINE__, FALSE, "in cfginput(), %s, for host %s, no input!", cfgfileprev ? cfgfileprev : cfgfilecurr, hostcurr);
					cfgcurr++;
					if ((cfgcurr > cfglast) /* || (cfgcurr > OBJCFG) */) {
						stroffset = 0;
						return EOF;
					} else {
						stroffset = 0;
						bof = TRUE;
						bon = FALSE;
						bol = FALSE;
						return '#';	/* dummy char */
					}
				}
			}
			sprintf(linenum, "%-6d\n", clp->lineno);
			cfp = cp = linenum;
		} else if (bol) {
			macro = FALSE;
			domcr = clp->domcr;
			cfp = cp = clp->line;
		}
	}

	while (TRUE) {
		if (!macro) {
			if (((c = *(cp+stroffset++)) == '=') && !escape && isalpha((int)*(cp+stroffset))) {
				offset_save = stroffset;
				i = 0;
				while (isalnum((int)(mc = *(cp+stroffset++))) || strchr("-_.",mc))
					i++;
				macroname = (char *) malloc(i*sizeof(char));
				stroffset = offset_save;
				i = 0;
				while (isalnum((int)(mc = *(cp+stroffset++))) || strchr("-_.",mc)) {
					*(macroname+i++) = mc;
				}
				*(macroname+i) = '\0';
				if (!(idp = lookupident(macroname, MCRCFG))) {
					if (domcr)
						logexit(PIKTLOG, ERR, __FILE__, __LINE__, FALSE, "in cfginput(), %s, for host %s, unknown macro %s", cfgfilecurr, hostcurr, macroname);
					else
						stroffset -= i+1;
				} else if ((idp->type == MCR) && domcr) {
					dfp = idp->definition;
					mcp = dfp->component;
					macroffset = 0;
					macro = TRUE;
					stroffset--;
				} else /* not a macro */ {
					stroffset -= i+1;
				}
				free(macroname);
			}
		}
		if (macro) {
			if ((c = *(mcp+macroffset++)) == '\0') {
				dfp = dfp->next;
				if (dfp) {
					mcp = dfp->component;
					macroffset = 0;
					return ' ';
				} else {
					macro = FALSE;
				}
			} else
				return c;
		} else {
			if ((c == '\\') && *(cp+stroffset) && (*(cp+stroffset) == '=')) {
				escape = TRUE;
				stroffset++;
				return '=';
			}
			break;
		}
	}

	escape = FALSE;

	if (c != '\n') {
		return c;
	} else {
		if (bof) {
			bof = FALSE;
			bon = TRUE;
		} else if (bon) {
			bon = FALSE;
			bol = TRUE;
		} else if (bol) {
			if (clp->next) {
				bon = TRUE;
				clp = clp->next;
				if (clp) {
					cfgfileprev = cfgfilecurr;
					cfgfilecurr = clp->file;
				}
			} else {
				cfgcurr++;
				bof = TRUE;
				bon = FALSE;
				bol = FALSE;
			}
		}
		stroffset = 0;
		return '\n';
	}

}

/* return whether the leading parenthesis in the supplied string is a
   logic conditional parenthesis, as opposed to an arithmetic parenthesis;
   basically, we count the number of left versus right parentheses, until
   we reach certain turning points, where if the number of counted left
   parentheses exceeds the right count, we know that we are entering a
   logical condition
 */
int
condparen (char *s)
{
	register int i;
	int lp, rp;
	register unsigned char c, nc;
	int done;

	done = FALSE;
	lp = rp = 0;
	i = 0;
	while ((c = *(s+i))) {
		switch(c) {
			case '(':
				lp++;
				break;
			case ')':
				rp++;
				if (lp == rp) {
					if (context == STRING)
						rp--;	/* signalling cond */
					done = TRUE;
				}
				break;
			case '<':
				nc = *(s+i+1);
				switch(nc) {
					case ' ':
					case '>':
					case '\n':
					case '\t':
						done = TRUE;
						break;
					default:
						break;
				}
				break;
			case '>':
				nc = *(s+i+1);
				switch(nc) {
					case ' ':
					case '\n':
					case '\t':
						done = TRUE;
						break;
					default:
						break;
				}
				break;
			case '-':	/* e.g., (-e $name) */
				nc = *(s+i+1);
				if (strchr("ezfdcbplS",nc) && (lp == 1))
					done = TRUE;
				break;
			default:
				break;
		}
		if (done) break;
		i++;
	}
	return lp > rp ? TRUE : FALSE;
}

/* when processing a statement like

      $substitute($line, "t", "X", 2) eq "this is a Xest"

   the lexer will eventually pattern match

      "t", "X", 2) eq "this is a Xest"

   this function isolates the next function string, i.e.,

      "t"

   returning its length, in this example, 3
 */
int
fnstr (char *s)
{
	register int i, j;
	register unsigned char c;
	register int q;
	int done;

	done = FALSE;
	q = 0;
	i = 0;
	while ((c = *(s+i))) {
		switch(c) {
			case '"':
				if (*(s+i+1) == '\0')
					done = TRUE;
				else
					q++;
				break;
			case '.':
			case ',':
			case ')':
				if (even(q)) {
					done = TRUE;
					j = 1;
					while ((c = *(s+i-j)) != '\"') {
						if (isspace(c))
							j++;
						else {
							done = FALSE;
							break;
						}
					}
					if (done)
						i -= j;
				}
				break;
			default:
				break;
		}
		if (done) break;
		i++;
	}
	return i;
}
