/*
   cmd.c -- PIKT piktc command-line processing routines

   $Header: /home/cvs/ucgsb/pikt/src/piktc/cmd.c,v 1.10 1999/04/14 12:54:32 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "common.h"

nam *purge(nam *, nam *, nam *, int *, int);
nam *expand(nam *);

/* check whether alerts, programs, objectsets & hosts specified in the
   command line are registered in the config files; also whether
   alert/program/objectset/host specifications are consistent with other
   command-line options; note that cmdparse() ensures that "all", if
   present, is the sole member of a list
 */
void
preprocesscmd (void)
{
	cfgln *clp;
	nam *nmp1, *nmp2;
	static char *name;
	char *cp;
	ident *idp;

	if (    ishelp
	     || (!incmdhosts && !excmdhosts)
	     || ((ischeck || isdoublecheck) && (incmdalerts || excmdalerts || incmdprograms || excmdprograms || incmdobjects || excmdobjects))
	     || ((!(ischeck || isdoublecheck) && !isrestart && !iskill) && (!incmdalerts && !excmdalerts && !incmdprograms && !excmdprograms && !incmdobjects && !excmdobjects))
	     || ((isdisable || isenable || isshow) && (!incmdalerts && !excmdalerts))
	     || ((isdelete || isinstall) && (!incmdalerts && !excmdalerts && !incmdprograms && !excmdprograms && !incmdobjects && !excmdobjects))
	   ) {
		bailout = TRUE;
		return;
	}
	foreach (clp, configs[ALTCFG])
		if (!isspace((int)(clp->line[0])) && (clp->line[0] != '#')) {
			name = strdup(clp->line);
			*(name+strlen(name)-1) = '\0';
			regname(&cfgalerts, name, (char *) NULL);
			free(name);
		}
	foreach (nmp1, incmdalerts) {
		if (!strcmp(nmp1->name,"all"))
			break;
		foreach (nmp2, cfgalerts)
			if (!strcmp(nmp1->name, nmp2->name))
				break;
		if (!nmp2) {
			fprintf(stderr, "unknown alert: %s\n", nmp1->name);
			nmp1->state = ERROR;
			bailout = TRUE;
		}
	}
	foreach (nmp1, excmdalerts) {
		if (!strcmp(nmp1->name,"all"))
			break;
		foreach (nmp2, cfgalerts)
			if (!strcmp(nmp1->name, nmp2->name))
				break;
		if (!nmp2) {
			fprintf(stderr, "unknown alert: %s\n", nmp1->name);
			nmp1->state = ERROR;
			bailout = TRUE;
		}
	}
	if (!bailout)
		cmdalerts = purge(cfgalerts, incmdalerts, excmdalerts, &alertsall, ALT);
	foreach (clp, configs[PRGCFG])
		if (!isspace((int)(clp->line[0])) && (clp->line[0] != '#')) {
			name = strdup(clp->line);
			*(name+strlen(name)-1) = '\0';
			if ((cp = strrchr(name, '/'))) {
				*cp = '\0';
				regname(&cfgprograms, cp+1, name);
			} else
				regname(&cfgprograms, name, (char *) NULL);
			free(name);
		}
	foreach (nmp1, incmdprograms) {
		if (!strcmp(nmp1->name,"all"))
			break;
		foreach (nmp2, cfgprograms)
			if (!strcmp(nmp1->name, nmp2->name))
				break;
		if (!nmp2) {
			fprintf(stderr, "unknown program: %s\n", nmp1->name);
			nmp1->state = ERROR;
			bailout = TRUE;
		}
	}
	foreach (nmp1, excmdprograms) {
		if (!strcmp(nmp1->name,"all"))
			break;
		foreach (nmp2, cfgprograms)
			if (!strcmp(nmp1->name, nmp2->name))
				break;
		if (!nmp2) {
			fprintf(stderr, "unknown program: %s\n", nmp1->name);
			nmp1->state = ERROR;
			bailout = TRUE;
		}
	}
	if (!bailout)
		cmdprograms = purge(cfgprograms, incmdprograms, excmdprograms, &programsall, PRG);
	foreach (clp, configs[OBJCFG])
		if (!isspace((int)(clp->line[0])) && (clp->line[0] != '#')) {
			name = strdup(clp->line);
			*(name+strlen(name)-1) = '\0';
			regname(&cfgobjects, name, (char *) NULL);
			free(name);
		}
	foreach (nmp1, incmdobjects) {
		if (!strcmp(nmp1->name,"all"))
			break;
		foreach (nmp2, cfgobjects)
			if (!strcmp(nmp1->name, nmp2->name))
				break;
		if (!nmp2) {
			fprintf(stderr, "unknown objectset: %s\n", nmp1->name);
			nmp1->state = ERROR;
			bailout = TRUE;
		}
	}
	foreach (nmp1, excmdobjects) {
		if (!strcmp(nmp1->name,"all"))
			break;
		foreach (nmp2, cfgobjects)
			if (!strcmp(nmp1->name, nmp2->name))
				break;
		if (!nmp2) {
			fprintf(stderr, "unknown objectset: %s\n", nmp1->name);
			nmp1->state = ERROR;
			bailout = TRUE;
		}
	}
	if (!bailout)
		cmdobjects = purge(cfgobjects, incmdobjects, excmdobjects, &objectsall, OBJ);
	foreach (nmp1, incmdhosts) {
		if (!strcmp(nmp1->name,"all"))
			break;
		if (!(idp = lookupident(nmp1->name, SYSCFG))) {
			fprintf(stderr, "unknown os, host, group, or alias: %s\n", nmp1->name);
			nmp1->state = ERROR;
			bailout = TRUE;
		} else if ((idp->type != OPS) && (idp->type != GRP) && (idp->type != HST) && (idp->type != ALS)) {
			fprintf(stderr, "%s is not an os, group, host, or alias\n", nmp1->name);
			nmp1->state = ERROR;
			bailout = TRUE;
		}
	}
	foreach (nmp1, excmdhosts) {
		if (!strcmp(nmp1->name,"all"))
			break;
		if (!(idp = lookupident(nmp1->name, SYSCFG))) {
			fprintf(stderr, "unknown os, host, group, or alias: %s\n", nmp1->name);
			nmp1->state = ERROR;
			bailout = TRUE;
		} else if ((idp->type != OPS) && (idp->type != GRP) && (idp->type != HST) && (idp->type != ALS)) {
			fprintf(stderr, "%s is not an os, group, host, or alias\n", nmp1->name);
			nmp1->state = ERROR;
			bailout = TRUE;
		}
	}
	if (!bailout)
		cmdhosts = purge(cfghosts, incmdhosts, excmdhosts, &hostsall, HST);
	if (    !cmdhosts
		/*
	     || ((ischeck || isdoublecheck) && (cmdalerts || cmdprograms || cmdobjects))
	     || (!(ischeck || isdoublecheck) && (!cmdalerts && !cmdprograms && !cmdobjects))
	     || ((isdisable || isenable || isshow) && !cmdalerts)
	     || ((isinstall || isdelete) && (!cmdalerts && !cmdprograms && !cmdobjects))
		 */
	   )
		bailout = TRUE;
	return;
}

/* register the command-line name in the specified list
 */
nam *
regname (nam **list, const char *name, char *dir)
{
	nam *nmp, *nnmp;

	if (*list) {
		foreach (nmp, *list) {
			if (!strcmp(nmp->name, name))
				return nmp;
		}
	}
	nnmp = (nam *) malloc(sizeof(nam));
	if (!nnmp)
		logexit(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in regname(), malloc() failure for nnmp");
	nnmp->next = (nam *) NULL;
	nnmp->name = strdup(name);
	if (dir)
		nnmp->aux = strdup(dir);
	else
		nnmp->aux = dir;
	nnmp->state = ADD;
	if (list)
		nnmp->next = *list;
	*list = nnmp;
	return *list;	
}

/* garbage collect list items
 */
void
retire (nam **names)
{
	nam *nmp;
	static nam *tnmp;

	nmp = *names;
	while (nmp) {
		tnmp = nmp;
		nmp = nmp->next;
		FREE(tnmp->name);
		FREE(tnmp->aux);
		FREE(tnmp);
	}
	*names = (nam *) NULL;
}

/* do a lookup on the names specified in the command line
 */
nam *
lookupname (nam *list, char *name)
{
	nam *nmp;

	foreach (nmp, list)
		if (!strcmp(name,nmp->name))
			break;
	return nmp;
}


/* given an include list, purge members in the given exclude list
 */
nam *
purge (nam *cfgitems, nam *incmditems, nam *excmditems, int *itemsall, int type)
{
	nam *initems, *exitems;
	static nam *nmp1, *nmp2;

	initems = exitems = (nam *) NULL;
	if (incmditems) {
		if (!strcmp(incmditems->name,"all")) {
			*itemsall = TRUE;
			initems = cfgitems;
		} else if (type == HST) {
			initems = expand(incmditems);
			if (!initems)
				return (nam *) NULL;
		} else
			initems = incmditems;
	}
	if (excmditems) {
		*itemsall = FALSE;
		if (!strcmp(excmditems->name,"all")) {
			exitems = cfgitems;
		} else if (type == HST) {
			exitems = expand(excmditems);
			if (!exitems)
				return (nam *) NULL;
		} else
			exitems = excmditems;
	}
	if (exitems && !initems)
		initems = cfgitems;
	/* foreach of the items in the exclude list,
	   seek out and purge from the include list
	 */
	foreach (nmp1, exitems) {
		if (!(nmp2 = initems))	/* if excluded initems down to nothing */
			break;
		if (!strcmp(nmp1->name, nmp2->name)) {
			initems = nmp2->next;
		} else {
			while (nmp2->next) {
				if (!strcmp(nmp1->name, nmp2->next->name)) {
					nmp2->next = nmp2->next->next;
					break;
				}
				nmp2 = nmp2->next;
			}
		}	
	}
	if (excmditems && strcmp(excmditems->name, "all"))
		retire(&exitems);
	return initems;
}

/* given a host/alias/group list, replace any host alias with the hostname
   it represents, and replace any group with the hostnames it represents;
   e.g., the list

      vienna rome db linux

   would expand to

      vienna milan athens2 athens4 milan seville plymouth

   if vienna were a plain hostname; rome were an alias for milan; db were
   a host group including athens2, athens4, and milan; and linux were an
   os group including seville and plymouth
 */
nam *
expand (nam *oldlist)
{
	nam *newlist;
	nam *nmp;
	ident *idp;
	unsigned long *hbs;
	unsigned long hbsl;
	register int i, j;

	newlist = (nam *) NULL;
	foreach (nmp, oldlist) {
		idp = lookupident(nmp->name, SYSCFG);
		switch (idp->type) {
			case HST:
				regname(&newlist, idp->name, (char *) NULL);
				break;
			case ALS:
				regname(&newlist, idp->alias, (char *) NULL);
				break;
			case GRP:
			case OPS:
				hbs = idp->bs;
				/* foreach unsigned long in the host bitset */
				for (i=0; i<hbslongs; i++) {
					hbsl = *(hbs+i);
					/* foreach bit in the current
					   unsigned long
					 */
					for (j=0; j<8*sizeof(unsigned long); j++) {
						if (hbsl & (1<<j))
							regname(&newlist, *(hosts+((i*sizeof(unsigned long)*8)+j)), (char *) NULL);
					}
				}
				break;
			default:
				fprintf(stderr, "%s is not an os, group, or host\n", nmp->name);
				bailout = TRUE;
				return (nam *) NULL;
		}
	}
	return newlist;
}
