/*
   piktc.c -- PIKT master control program, and config file compiler

   $Header: /home/cvs/ucgsb/pikt/src/piktc/piktc.c,v 1.12 1999/04/15 04:25:05 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA

   --------------------------------------------------------------------------

   Usage:  piktc <-cCdeGhiklrstTv>
                 [+|-D        <defines>]
                 [+|-A[f] all|<alerts>]
                 [+|-O[f] all|<objects>]
                 [+|-P[f] all|<programs>]
                  +|-H all|<hosts>
                 [ALL]

          -c                     syntax check all config files
          -C                     syntax doublecheck all config files
          -d                     disable alert(s)
          -e                     enable alert(s)
          -G                     run in debug mode
          -h                     display program help
          -i                     install alert/program/object file(s)
          -k                     kill alert daemon (piktd)
          -l                     list alert/program/object file(s)
          -r                     (re)start alert daemon (piktd)
          -s                     show alert(s) status
          -t                     delete alert/program/object/history/log
                                   file(s)
          -T                     run in test mode
          -v                     isverbose
          +|-D        <defines>  define/undefine defines
          +|-A[f] all|<alerts>   include/exclude (fix) alerts
          +|-O[f] all|<objects>  include/exclude (fix) object files
          +|-P[f] all|<programs> include/exclude (fix) program files
          +|-H    all|<hosts>    include/exclude hosts, oses, groups, aliases
          ALL                    +A all +O all +P all +H all

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "common.h"

const char piktc_version[] = "PIKT piktc version " VERSION;

/* command-line options */
int ischeck = FALSE;
int isdebug = FALSE;
int isdelete = FALSE;
int isdoublecheck = FALSE;
int isdisable = FALSE;
int isenable = FALSE;
int ishelp = FALSE;
int isinstall = FALSE;
int iskill = FALSE;
int islist = FALSE;
int isrestart = FALSE;
int istest = FALSE;
int isshow = FALSE;
int isverbose = FALSE;

int prescan;
int postscan;

int bailout = FALSE;	/* signifies command-line error,
			   or error in systems.cfg
			 */

int uid = 0;		/* default to root */
char *master;
char *domain;
char *accessmode;

char hostname[MAXIDENT+1];
char domainname[MAXIDENT+1];

char **argvcurr;	/* current command-line argument */
char **argvlast;	/* last command-line argument */

unsigned stroffset = 0;

int lineno = 0;

int cfgerr = FALSE;
int stzerr = FALSE;

size_t hbslongs;
size_t dbslongs;

int hostnum = 0;
int opsysnum = 0;
int defnum = 0;

char **hosts;
char *hostcurr;

char **opsys;

unsigned long *hbscurr = (unsigned long *) NULL;

/* cmdalerts/cmdprograms/cmdobjects/cmdhosts are the
   alerts/programs/objectsets/hosts specified in the command line

   cfgalerts/cfgprograms/cfgobjects are the alerts/programs/objectsets
   included in the tagged config files

   alerts/programs/objects are the alerts/programs/objectsets being written
   for the current host in the current program run

   since we bailout if any alert/program/objectsset (or host) specified in
   the command line is bogus (meaning it's not found in the config files),
   cmdalerts/cmdprograms/cmdobjects is a subset of
   cfgalerts/cfgprograms/cfgobjects

   alerts/programs/objects is identical to cmdalerts/cmdprograms/cmdobjects
   if comand-line alerts/programs/objects are specified, else
   cmdalerts/cmdprograms/cmdobjects is empty and alerts/programs/objects
   are the things actually written
 */

nam *alerts,   *cfgalerts,   *incmdalerts,   *excmdalerts,   *cmdalerts,
    *programs, *cfgprograms, *incmdprograms, *excmdprograms, *cmdprograms,
    *objects,  *cfgobjects,  *incmdobjects,  *excmdobjects,  *cmdobjects,
               *cfghosts,    *incmdhosts,    *excmdhosts,    *cmdhosts;

int alertsall, programsall, objectsall, hostsall;	/* was "all" specified
							   on command line?
							 */

FILE	*stripin,	*stripout,
	*tagin,		*tagout,
	*sysin,		*sysout,
	*cfgin,		*cfgout;

/* the order of the following .cfg files is vitally important! */
const char *cfgfiles[] = { "systems.cfg",
                           "defines.cfg",
                           "macros.cfg",
                           "alarms.cfg",
                           "alerts.cfg",
                           "programs.cfg",
                           "objects.cfg"
                         };

char *cfgfilecurr = (char *) NULL;
char *cfgfileprev = (char *) NULL;

char *infile = (char *) NULL;

int cfgcurr;
int cfglast;

cfgln *configs[OBJCFG+1];	/* assumes OBJCFG is last cfg file */

/* the order of the following is vitally important! */
const char *cfgbof[] = {       "/\n",		/* SYSCFG */
                              "//\n",		/* DEFCFG */
                             "///\n",		/* MCRCFG */
                            "////\n",		/* ALMCFG */
                           "/////\n",		/* ALTCFG */
                          "//////\n",		/* PRGCFG */
                         "///////\n"		/* OBJCFG */
                       };

char *cfp;

cfgln *clplast = (cfgln *) NULL;

unsigned bof;
unsigned bon;
unsigned bol;

def *defcurr;

char *mins, *hrs, *dom, *mons, *dow;
char *priority;
char *nicecmd, *mailcmd, *lpcmd;

int stripalm, stripprg, stripobj;

int domacros;
int dostrip;

void preprocess(int, int, int);
void processcmd(int, char **);
void process(unsigned long*, int, char **);
void dodoublecheck(void);
int chkpiktc_svc(char *host);

int
main (int argc, char *argv[])
{
	struct stat fbuf;

	init(argc, argv, (void (*)(void))piktc_usage);

	alerts = cfgalerts = incmdalerts = excmdalerts = cmdalerts =
	objects = cfgobjects = incmdobjects = excmdobjects = cmdobjects =
	programs = cfgprograms = incmdprograms = excmdprograms = cmdprograms =
	          cfghosts = incmdhosts = excmdhosts = cmdhosts = (nam *)NULL;
	stripalm = stripprg = stripobj = TRUE;

	argvcurr = argv+1;	/* skip the program name */
	argvlast = argv+argc-1;
	postscan = FALSE;
	cmdparse();		/* further parse the command line */

	/* apply the lock, to prevent other concurrent piktc sessions */
	if (islock(piktclock)) {
		fprintf(stderr, "%s found!  Another piktc session may be in progress.\nTry again later (else rm the file).\n", piktclock);
		fflush(stderr);
		exit(0);
	} else {
		if (mklock(piktclock, (char *) NULL, TRUE, TRUE) == ERR)
			logexit(PIKTLOG, ERR, __FILE__, __LINE__, FALSE, "in main(), mklock() error for %s", piktclock);
	}

	/* preprocess the config files; the order is vitally important! */
	preprocess(SYSCFG, TRUE, TRUE);
	preprocess(SYSCFG, TRUE, FALSE);

	/* only preprocess the defines.cfg file if it exists
	 */
	cfgfilecurr = infile = (char *) malloc((strlen(cfgdir)+strlen(cfgfiles[DEFCFG])+2) * sizeof(char));
	cfgfileprev = (char *) NULL;
	sprintf(infile, "%s/%s", cfgdir, cfgfiles[DEFCFG]);
	if (stat(infile, &fbuf) == 0) {
		preprocess(DEFCFG, TRUE, TRUE);
		preprocess(DEFCFG, TRUE, FALSE);
	}

	preprocess(MCRCFG, TRUE, FALSE);
	preprocess(ALMCFG, stripalm, FALSE);
	preprocess(ALTCFG, TRUE, FALSE);
	preprocess(PRGCFG, stripprg, FALSE);
	preprocess(OBJCFG, stripobj, FALSE);

	preprocesscmd();	/* preprocess command-line items */
	
	if (bailout)
		piktc_usage();
	else
		processcmd(argc, argv);		/* execute the command(s) */

	/* remove the lock */
	if (rmlocks(TRUE, TRUE) == ERR)
		logexit(PIKTLOG, ERR, __FILE__, __LINE__, FALSE, "in main(), rmlocks() error");

	if (bailout)
		exit(1);
	else
		exit(0);
}

/* output program usage to screen
 */
void
piktc_usage (void)
{
	printf("\nUsage:  piktc <-cCdeGhiklrstTv>\n\t      [+|-D        <defines>]\n\t      [+|-A[f] all|<alerts>]\n\t      [+|-O[f] all|<objects>]\n\t      [+|-P[f] all|<programs]\n\t       +|-H    all|<hosts>\n\t      [ALL]\n\n");
	printf("\t-c\t\t\tsyntax check all config files\n");
	printf("\t-C\t\t\tsyntax doublecheck all config files\n");
	printf("\t-d\t\t\tdisable alert(s)\n");
	printf("\t-e\t\t\tenable alert(s)\n");
	printf("\t-G\t\t\trun in debug mode\n");
	printf("\t-h\t\t\tdisplay program help\n");
	printf("\t-i\t\t\tinstall alert/program/object file(s)\n");
	printf("\t-k\t\t\tkill alert daemon (piktd)\n");
	printf("\t-l\t\t\tlist alert/program/object file(s)\n");
	printf("\t-r\t\t\t(re)start alert daemon (piktd)\n");
	printf("\t-s\t\t\tshow alert(s) status\n");
	printf("\t-t\t\t\tdelete alert/program/object/history/log\n\t\t\t\t  file(s)\n");
	printf("\t-T\t\t\trun in test mode\n");
	printf("\t-v\t\t\tverbose\n");
	printf("\t+|-D        <defines>\tdefine/undefine defines\n");
	printf("\t+|-A[f] all|<alerts>\tinclude/exclude (fix) alerts\n");
	printf("\t+|-O[f] all|<objects>\tinclude/exclude (fix) object files\n");
	printf("\t+|-P[f] all|<programs>\tinclude/exclude (fix) program files\n");
	printf("\t+|-H    all|<hosts>\tinclude/exclude hosts, oses, groups, aliases\n");
	printf("\tALL\t\t\t+A all +O all +P all +H all\n\n");
	return;
}

/* for each host in the host-bit set, process the piktc command: enable/
   disable alerts, install/delete files, list files, kill/restart alert
   daemon, show alert status, check config file validity
 */
void
processcmd (int argc, char **argv)
{
	ident *idp;
	nam *hstp;
	char cmd[MAXCMD+1];
	static unsigned long *hbs;
	static unsigned long *tbs;
	unsigned long hbsl;
	register int i, j;

	/* for each host in the command-line set, turn on its corresponding
	   bit in the host bitset
	 */
	hbs = (unsigned long *) calloc(hbslongs, sizeof(unsigned long));
	foreach (hstp, cmdhosts) {
		idp = lookupident(hstp->name, SYSCFG);
		if (idp) {
			tbs = hbs;
			hbs = bsunion(hbs, idp->bs, hbslongs);
			free(tbs);
		} else	/* validatecmditems() should have caught this */
			fprintf(stderr, "unknown host or group:  %s\n", hstp->name);
			fflush(stderr);
	}
	/* foreach unsigned long in the host bitset */
	for (i=0; i<hbslongs; i++) {
		hbsl = *(hbs+i);
		/* foreach bit in the current unsigned long */
		for (j=0; j<8*sizeof(unsigned long); j++) {
			if (isdebug)
				logreturn(PIKTLOG, DEBUG, __FILE__, __LINE__, FALSE, "i is %d, j is %d, 1<<j is %x, hbsl is %x", i, j, 1<<j, hbsl);
			/* if the bit for this host is on */
			if (hbsl & (1<<j)) {
				hostcurr = *(hosts+((i*sizeof(unsigned long)*8)+j));
				FREE(hbscurr);
				hbscurr = bsdup(getbs(hostcurr, HST), hbslongs);
				cfgerr = FALSE;
				stzerr = FALSE;

				if (isverbose)
					printf("\n%s%sing %s...\n", isdoublecheck ? "double" : "", (ischeck || isdoublecheck) ? "check" : "process", hostcurr);

				/* if not doing a syntax check, determine if
				   remote piktc_svc is running; if not, say
				   so, and bypass the current host
				 */
				if (! (ischeck || isdoublecheck)) {
					if (chkpiktc_svc(hostcurr) != 0) {
						if (isverbose) {
							fprintf(stderr, "remote piktc_svc not running\n");
							fflush(stderr);
						}
						continue;
					}
				}

				/* disable remote alerts */
				if ((/* isinstall || */ isdisable || isenable || isdelete) && cmdalerts)
					disablealerts(hostcurr, alertsall ? cfgalerts : cmdalerts, isverbose && !isenable /* && !isinstall */);

				/* delete remote alert files */
				if ((/* isinstall || */ isdelete) && cmdalerts)
					deletefiles(hostcurr, alertsall ? cfgalerts : cmdalerts, altdir, ".alt", isverbose && !isinstall);

				/* delete remote history/log files */
				if (isdelete && cmdalerts) {
					deletefiles(hostcurr, alertsall ? cfgalerts : cmdalerts, hstdir, ".hst", isverbose && !isinstall);
					deletefiles(hostcurr, alertsall ? cfgalerts : cmdalerts, logdir, ".log", isverbose && !isinstall);
				}

				/* delete remote object files */
				if ((/* isinstall || */ isdelete) && cmdobjects)
					deletefiles(hostcurr, objectsall ? cfgobjects : cmdobjects, objdir, ".obj", isverbose && !isinstall);

				/* install/check/doublecheck cfg/dat files */
				if (isinstall || isenable || ischeck || isdoublecheck || (isdelete && cmdprograms)) {

					/* clear out staging dir */
					sprintf(cmd, "/bin/rm -f %s/*", stgdir);
					system(cmd);

					/* process config files for current host */
					process(hbscurr, argc, argv);

					/* garbage collect */
					cleanup();

					if (isdoublecheck) {
						dodoublecheck();
					}

					if (cfgerr || ischeck || isdoublecheck)
						continue;

				}

				/* note: it is necessary to delete program
				   files after invoking process(), for we
				   need to know the programs list for the
				   current machine; otherwise, for programs
				   installed outside the default PIKT program
				   directory and inside a directory NFS
				   cross-mounted to more than one machine,
				   we risk blowing them away by passing
				   cfgprograms to deletefiles() (in the
				   manner we do above) when processing
				   a later machine
				 */

				/* delete remote program files */
				if ((/* isinstall || */ isdelete) && cmdprograms)
					deletefiles(hostcurr, programs, prgdir, "", isverbose && !isinstall);

				/* note: an alert/program/object may be
				   specified in the command line but not be
				   written to the staging directory, because
				   it doesn't apply to a particular host,
				   which is why we can't install/enable
				   cmdalerts/cmdprograms/cmdobjects, which
				   might include alerts/programs/objects
				   not processed
				 */

				/* install remote alert files */
				if (isinstall && cmdalerts)
					installfiles(hostcurr, alerts, altdir, ".alt", (mode_t)(S_IRUSR|S_IWUSR|S_IRGRP), isverbose);

				/* install remote program files */
				if (isinstall && cmdprograms)
					installfiles(hostcurr, programs, prgdir, "", (mode_t)(S_IRWXU|S_IRGRP|S_IXGRP), isverbose);

				/* install remote object files */
				if (isinstall && cmdobjects)
					installfiles(hostcurr, objects, objdir, ".obj", (mode_t)(S_IRUSR|S_IWUSR|S_IRGRP), isverbose);

				/* enable remote alerts */
				if ((/* isinstall || */ isenable) && cmdalerts)
					enablealerts(hostcurr, alertsall ? alerts : alerts, isverbose);

				/* list remote alert files */
				if (islist && cmdalerts)
					listfiles(hostcurr, alertsall ? cfgalerts: cmdalerts, altdir, ".alt", isverbose);

				/* list remote program files */
				if (islist && cmdprograms)
					listfiles(hostcurr, programsall ? cfgprograms : cmdprograms, prgdir, "", isverbose);

				/* list remote object files */
				if (islist && cmdobjects)
					listfiles(hostcurr, objectsall ? cfgobjects : cmdobjects, objdir, ".obj", isverbose);

				/* kill remote alert daemon */
				if (iskill | isrestart)
					killdaemon(hostcurr, isverbose && !isrestart);

				/* (re)start remote alert daemon */
				if (isrestart) {
					sleep(2);	/* give time for daemon to die from just-concluded kill */
					restartdaemon(hostcurr, isverbose);
				}

				/* show remote alert status */
				if (isshow && cmdalerts)
					showstata(hostcurr, alertsall ? cfgalerts : cmdalerts, isverbose);

				retire(&alerts);
				retire(&programs);
				retire(&objects);
			}
		}
	}
	free(hbs);
	return;
}

/* open the config file, filter out C++ style comments in strip, and
   pass the output to the sys parser for systems.cfg, and to the
   cfg parser for all other .cfg files; after preprocessing, we no
   longer need refer to config files on disk, as all information is
   stored in internal data structures
 */
void
preprocess (int cfgtype, int strip, int isprescan)
{
	int fd[2];
	pid_t pid;
	struct stat fbuf;

	cfgcurr = cfgtype;
	dostrip = strip;
	prescan = isprescan;
	if (! isprescan) {
		switch (cfgtype) {
			case SYSCFG:
				hbslongs = (int)((hostnum-1)/(8*sizeof(long))) + 1;
				hosts = (char **) malloc(hostnum*(sizeof(char *)));
				opsys = (char **) malloc(opsysnum*(sizeof(char *)));
				hostnum = 0;
				opsysnum = 0;
				/* currently serve no purpose
				hbsfull = (unsigned long *) calloc(hbslongs, sizeof(unsigned long));
				hbsempty = (unsigned long *) calloc(hbslongs, sizeof(unsigned long));
				 */
				break;
			case DEFCFG:
				dbslongs = (int)((defnum-1)/(8*sizeof(long))) + 1;
				gdbs = (unsigned long *) calloc(dbslongs, sizeof(unsigned long));
				gubs = (unsigned long *) calloc(dbslongs, sizeof(unsigned long));
				/* currently serve no purpose
				dbsfull = (unsigned long *) calloc(dbslongs, sizeof(unsigned long));
				dbsempty = (unsigned long *) calloc(dbslongs, sizeof(unsigned long));
				 */
				break;
			default:
				break;
		}
	}
	if (!prescan)
		domacros = TRUE;
	incfiles = (void *) NULL;
	incnum = 0;
	lineno = 1;
	/* linux chokes on this
	FREE(infile);
	 */
	cfgfilecurr = infile = (char *) malloc((strlen(cfgdir)+strlen(cfgfiles[cfgtype])+2) * sizeof(char));
	cfgfileprev = (char *) NULL;
	sprintf(infile, "%s/%s", cfgdir, cfgfiles[cfgtype]);
	if (stat(infile, &fbuf) != 0) {
		logexit(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in preprocess(), stat() error for %s", infile);
	}
	if (dostrip) {
		if (pipe(fd) < 0)
			logexit(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in preprocess(), pipe() error");
		switch (pid = fork()) {
			case -1:	/* error */
			logexit(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in preprocess(), fork() error");
				break;
			/* pipe from child to parent */
			case 0:		/* child */
				close(fd[0]);
				stripin = fopen(infile, "r");
				stripout = fdopen(fd[1], "w");
				striplex();
				fclose(stripin);
				fclose(stripout);
				close(fd[1]);
				exit(0);
				break;
			default:	/* parent */
				close(fd[1]);
				#if defined(TIME_TO_SLEEP)
				sleep(TIME_TO_SLEEP);
				#endif
				switch (cfgtype) {
					case SYSCFG:
						sysin = fdopen(fd[0], "r");
						sysparse();
						fclose(sysin);
						break;
					default:
						tagin = fdopen(fd[0], "r");
						tagparse();
						fclose(tagin);
						break;
				}
				close(fd[0]);
				break;
		}
	} else {
		/* for now, only ALMCFG/PRGCFG/OBJCFG are potentially
		   not stripped, SYSCFG is always stripped, so switch
		   not really necessary here
		 */
		switch (cfgtype) {
			case SYSCFG:
				sysin = fopen(infile, "r");
				sysparse();
				fclose(sysin);
				break;
			default:
				tagin = fopen(infile, "r");
				tagparse();
				fclose(tagin);
				break;
		}
	}
	return;
}

/* foreach macros, alarms, alerts, programs & objects config file, process
   the config lines; foreach config line, if its tag matches for the
   current host, then input it to the cfg parser
 */
void
process (unsigned long *hbs, int argc, char **argv)
{
	struct stat fbuf;

	#if DEEPDEBUG
	if (isdebug)
		dumpcfgs(hbs);
	#endif

	/* only call defparse() if there is a defines.cfg file
	 */
	cfgfilecurr = infile = (char *) malloc((strlen(cfgdir)+strlen(cfgfiles[DEFCFG])+2) * sizeof(char));
	cfgfileprev = (char *) NULL;
	sprintf(infile, "%s/%s", cfgdir, cfgfiles[DEFCFG]);
	if (stat(infile, &fbuf) == 0) {
		empty(gdbs, dbslongs);
		empty(gubs, dbslongs);
		defnum = 0;
		stroffset = 0;
		cfgcurr = DEFCFG;
		cfglast = DEFCFG;
		bof = TRUE;
		defparse();
		if (bailout) {	/* set in deferror() */
			if (rmlocks(TRUE, TRUE) == ERR)
				logreturn(PIKTLOG, ERR, __FILE__, __LINE__, FALSE, "in process(), rmlocks() error");
			exit(1);
		}
		/* now go back and apply any command-line def overrides */
		argvcurr = argv+1;	/* skip the program name */
		argvlast = argv+argc-1;
		postscan = TRUE;
		cmdparse();
		stroffset = 0;
		cfgcurr = MCRCFG;
		cfglast = OBJCFG;
		bof = TRUE;
		defparse();
		#if DEEPDEBUG
		/* only dump a second time after the lines are marked
		   for doproc
		 */
		if (isdebug)
			dumpcfgs(hbs);
		#endif
		if (bailout) {	/* set in deferror() */
			if (rmlocks(TRUE, TRUE) == ERR)
				logreturn(PIKTLOG, ERR, __FILE__, __LINE__, FALSE, "in process(), rmlocks() error");
			exit(1);
		}
	}

	stroffset = 0;
	cfgcurr = MCRCFG;
	cfglast = OBJCFG;
	cfgfilecurr = cfgfiles[MCRCFG];
	cfgfileprev = (char *) NULL;
	bof = TRUE;
	bon = FALSE;
	bol = FALSE;
	cfgparse();
	return;
}

/* do a second syntax check on each alert in STGDIR, by running them through
   a pikt -C check (which does a more thorough syntax check than piktc);
   if problems are discovered in the pikt doublecheck, it might be nice to
   save each specimen for later inspection, but we'll reserve implementation
   of that for another time!
 */
void
dodoublecheck (void)
{
	nam *nmp;
	static char *alt;
	char cmd[MAXCMD+1];
	struct stat fbuf;

	foreach (nmp, alerts) {
		if (nmp->name) {
			FREE(alt);
			alt = (char *) malloc((strlen(stgdir)+strlen(nmp->name)+6) * sizeof(char));
			sprintf(alt, "%s/%s.alt", stgdir, nmp->name);
			if (lstat(alt, &fbuf) == 0) {	/* alert file found */
				sprintf(cmd, "%s/pikt -C +A %s", bindir, nmp->name);
				system(cmd);
			}
		}
	}
	FREE(alt);
	return;
}

/* check to see if piktc_svc on host is up; return 0 if so, ERROR if down
 */
int
chkpiktc_svc (char *host)
{
	char cmd[MAXCMD+1];
	char line[MAXLINE+1];
	FILE *rpcinfo;

	#if defined(RPCINFOCMD_1)
	sprintf(cmd, "%s -T udp %s %lu 2>&1", RPCINFOCMD, host, (unsigned long) RPCPRGNUM);
	#else
	sprintf(cmd, "%s -u %s %lu 2>&1", RPCINFOCMD, host, (unsigned long) RPCPRGNUM);
	#endif
	if ((rpcinfo = popen(cmd, "r")) == NULL) {
		logreturn(PIKTLOG, ERR, __FILE__, __LINE__, FALSE, "in chkpiktc_svc(), popen() error");
		return ERROR;
	}
	fgets(line, MAXLINE+1, rpcinfo);	/* just get the first line */
	pclose(rpcinfo);
	if (strstr(line, "ready and waiting"))
		return 0;
	else
		return ERROR;
}
