/*
   piktc_svc.c -- PIKT master control program (piktc) service daemon

   $Header: /home/cvs/ucgsb/pikt/src/piktc/piktc_svc.c,v 1.4 1999/04/13 01:03:42 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA

   --------------------------------------------------------------------------

   Usage:  piktc_svc [-GhT]

           -G              run in debug mode
           -h              show program help
           -T              run in test mode

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "common.h"

const char piktc_version[] = "PIKT piktc_svc version " VERSION;

/* command-line options */
/* int ischeck = FALSE; */
int isdebug = FALSE;
/* int isdelete = FALSE; */
/* int isdoublecheck = FALSE; */
/* int isdisable = FALSE; */
/* int isenable = FALSE; */
int ishelp = FALSE;
/* int isinstall = FALSE; */
/* int iskill = FALSE; */
/* int islist = FALSE; */
/* int isrestart = FALSE; */
int istest = FALSE;
/* int isshow = FALSE; */
/* int isverbose = FALSE; */

int uid = 0;		/* default to root */
char *master;
char *domain;
char *accessmode;

char hostname[MAXIDENT+1];
char domainname[MAXIDENT+1];

int
main (int argc, char *argv[])
{
	init(argc, argv, (void (*)(void))piktc_svc_usage);

	/* apply the lock, to prevent other concurrent piktc_svc runs */
	if (islock(piktc_svclock)) {
		fprintf(stderr, "%s found!  Another piktc session may be in progress.\nTry again later (else rm the file).\n", piktc_svclock);
		fflush(stderr);
		exit(0);
	} else {
		if (mklock(piktc_svclock, (char *) NULL, TRUE, TRUE) == ERR)
			logexit(PIKTLOG, ERR, __FILE__, __LINE__, FALSE, "in main(), mklock() error for %s", piktc_svclock);
	}

	logreturn(PIKTLOG, INFO, __FILE__, __LINE__, FALSE, "daemon (re)start");

	piktc_svc();	/* go into daemon mode */

	exit(0);	/* never reached */
}

/* output program usage to screen
 */
void
piktc_svc_usage (void)
{
	printf("\nUsage:  piktc_svc [-GhT]\n\n");
	printf("\t-G\t\t\trun in debug mode\n");
	printf("\t-h\t\t\tdisplay program help\n");
	printf("\t-T\t\t\trun in test mode\n\n");
	return;
}
