/*
   rpcsvc.c -- PIKT rpc service (piktc_svc) routines

   $Header: /home/cvs/ucgsb/pikt/src/piktc/rpcsvc.c,v 1.17 1999/04/14 14:21:28 harlan Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define PORTMAP
#include <rpc/rpc.h>

#if defined(HAVE_NETCONFIG_H)
#include <netconfig.h>
#endif

#include <rpc/xdr.h>
#include <rpc/clnt.h>
#include "rpc.h"

#include <sys/socket.h>
/*
#include <netinet/in.h>
#include <arpa/inet.h>
*/
#include <netdb.h>

#include <sys/utsname.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <dirent.h>

#if defined(HAVE_SYS_SYSTEMINFO_H)
	#include <sys/systeminfo.h>
#endif

#include <errno.h>

#include "common.h"

#ifdef NO_RPC_SVC
#define FILEEXIST_1_SVC		fileexist_1
#define INSTALLFILE_1_SVC	installfile_1
#define DELETEFILE_1_SVC	deletefile_1
#define EDITCONF_1_SVC		editconf_1
#define KILLDAEMON_1_SVC	killdaemon_1
#define RESTARTDAEMON_1_SVC	restartdaemon_1
#define SHOWSTATUS_1_SVC	showstatus_1
#else
#define FILEEXIST_1_SVC		fileexist_1_svc
#define INSTALLFILE_1_SVC	installfile_1_svc
#define DELETEFILE_1_SVC	deletefile_1_svc
#define EDITCONF_1_SVC		editconf_1_svc
#define KILLDAEMON_1_SVC	killdaemon_1_svc
#define RESTARTDAEMON_1_SVC	restartdaemon_1_svc
#define SHOWSTATUS_1_SVC	showstatus_1_svc
#endif

CLIENT *clnt;

pid_t daemonpid(void);

/* for all below, return 0 on success, else ERROR, NONE, DENIED, ...;
   log all errors including errno info, if available, in the local piktc.log
 */

/* attempt to authenticate the requesting piktc client; compare the
   client name to the approved master, if they don't match, deny the
   request; moreover, if access is set to local, check to see that the
   client's network domain matches the local network domain; if they
   don't match, deny the request; finally, check to see that the
   requesting account is the root account; if it isn't, deny the request
 */
int
chkrequest (struct svc_req *rqstp, const char *request)
{
	int res;
	char *cp;
	struct sockaddr_in *sock_in;
	unsigned long addr;
	struct hostent *hp;
	char h_name[MAXIDENT+1];
	struct authunix_parms *unix_cred;
	char *clntname;
	int clntuid;
	int authenticated;

	res = 0;
	authenticated = FALSE;

	clntname = "<<ERR>>";
	clntuid = -1;

	/* get the client host's (i.e., the would-be control machine's)
	   fully qualified domain name
	 */
	sock_in = svc_getcaller(rqstp->rq_xprt);
	hp = gethostbyaddr((char *) &sock_in->sin_addr, sizeof(sock_in->sin_addr), AF_INET);
	strcpy(h_name, hp->h_name);
	cp = strchr(h_name, '.');
	if (!cp) {	/* we don't have a fully qualified host domainname */
		/* look for the fully qualified domain name as the first
		   item in the h_aliases list; until we learn more about
		   the nature of this h_aliases list, i.e., how to proceed
		   to the next item, and when to stop, the first item is
		   as far as we go
		 */
		if (*(hp->h_aliases)) {
			strcpy(h_name, *(hp->h_aliases));
			cp = strchr(h_name, '.');
		}
	}

	/* compare the calling client's name against master,
	   possibly setting authenticated to TRUE
	 */
	if (!strcmp(h_name, master)) {	/* we have a match */
		authenticated = TRUE;
	} else {			/* no match */
		if (istest) {
			if (strstr(h_name, master))
				authenticated = TRUE;
		} else {
			if (!cp) {
				/* the h_name is not a fully qualified host
				   domain name, so cannot match master; we will
				   make one last attempt to authenticate this
				   client
				 */
			/* not sure about this one, so leave this out for now
				hp = gethostbyname(master);
				if (*(hp->h_addr_list)) {
					if (!strcmp((char *) &sock_in->sin_addr, *(hp->h_addr_list)))
						authenticated = TRUE;
				}
			 */
				;
			}
		}
	}

	if (!strcmp(accessmode, "local")) {
		/* restrict requests to within the local domain only,
		   possibly setting authenticated to FALSE
		   !cp means we don't have the client's domain
		 */
		if (!strcmp(domain, "domain")) {
			if (!cp || (!strstr(cp+1, domainname) && !strstr(domainname, cp+1)))
				authenticated = FALSE;
		} else {
			if (!cp || strcmp(cp+1, domain))
				authenticated = FALSE;
		}
	} /* else accessmode is "global", so don't check anything */

	addr = ntohl(sock_in->sin_addr.s_addr);

	if (authenticated) {
		unix_cred = (struct authunix_parms *) rqstp->rq_clntcred;
		clntname = unix_cred->aup_machname;
		clntuid = unix_cred->aup_uid;
		#ifdef SECURE
		/* this clntname check is somewhat redundant, given the
		   earlier check of h_name
		 */
		if (!clntname || !strlen(clntname) || strncmp(clntname, master, strlen(clntname))) {
			logreturn(PIKTLOG, ERR, __FILE__, __LINE__, FALSE, "in chkrequest(), %s request from uid %d on unauthorized client %s (addr %x)", request, clntuid, h_name, addr);
			res = DENIED;
		}
		#endif
		if (!istest && (clntuid != uid)) {
			logreturn(PIKTLOG, ERR, __FILE__, __LINE__, FALSE, "in chkrequest(), %s request from unauthorized uid %d on client %s (addr %x)", request, clntuid, h_name, addr);
			res = DENIED;
		}
	} else {
		logreturn(PIKTLOG, ERR, __FILE__, __LINE__, FALSE, "in chkrequest(), %s request from uid %d on unauthorized client %s (addr %x)", request, clntuid, h_name, addr);
		res = DENIED;
	}

	if (res != DENIED)
		logreturn(PIKTLOG, INFO, __FILE__, __LINE__, FALSE, "%s request from uid %d on client %s (addr %x)", request, clntuid, h_name, addr);

	if (isdebug) {
		logreturn(PIKTLOG, DEBUG, __FILE__, __LINE__, FALSE, "uid is %d", uid);
		logreturn(PIKTLOG, DEBUG, __FILE__, __LINE__, FALSE, "master is %s", master);
		logreturn(PIKTLOG, DEBUG, __FILE__, __LINE__, FALSE, "domain is %s", domain);
		logreturn(PIKTLOG, DEBUG, __FILE__, __LINE__, FALSE, "access is %s", accessmode);
		logreturn(PIKTLOG, DEBUG, __FILE__, __LINE__, FALSE, "hostname is %s", hostname);
		logreturn(PIKTLOG, DEBUG, __FILE__, __LINE__, FALSE, "domainname is %s", domainname);
		logreturn(PIKTLOG, DEBUG, __FILE__, __LINE__, FALSE, "clntuid is %d", clntuid);
		logreturn(PIKTLOG, DEBUG, __FILE__, __LINE__, FALSE, "client addr (sock_in->sin_addr.s_addr) is %x", sock_in->sin_addr.s_addr);
		logreturn(PIKTLOG, DEBUG, __FILE__, __LINE__, FALSE, "client addr (addr) is %x", addr);
		logreturn(PIKTLOG, DEBUG, __FILE__, __LINE__, FALSE, "client hostname (h_name) is %s", h_name);
		if (clntname)
			logreturn(PIKTLOG, DEBUG, __FILE__, __LINE__, FALSE, "client name (clntname) is %s", clntname);
		if (cp)
			logreturn(PIKTLOG, DEBUG, __FILE__, __LINE__, FALSE, "client domain is %s", cp+1);
		else
			logreturn(PIKTLOG, DEBUG, __FILE__, __LINE__, FALSE, "client domain is unknown");
	}

	return res;
}

/* report if a file exists
 */
int *
FILEEXIST_1_SVC (rfil *rf, struct svc_req *rqstp)
{
	static char *fn;
	struct stat fbuf;
	static int res;

	if ((res = chkrequest(rqstp, "fileexist")) != 0)
		return &res;

	fn = (char *) malloc((strlen(rf->directory)+strlen(rf->name)+strlen(rf->extension)+2) * sizeof(char));
	sprintf(fn, "%s/%s%s", rf->directory, rf->name, rf->extension);
	res = stat(fn, &fbuf);	/* stat returns 0 on success,
				   -1 (ERROR) on failure
				 */
	free(fn);
	return &res;
}

/* install a file
 */
int *
INSTALLFILE_1_SVC (rfil *rf, struct svc_req *rqstp)
{
	static char *fn;
	FILE *fp;
	static int res;

	if ((res = chkrequest(rqstp, "installfile")) != 0)
		return &res;

	fn = (char *) malloc((strlen(rf->directory)+strlen(rf->name)+strlen(rf->extension)+2) * sizeof(char));
	sprintf(fn, "%s/%s%s", rf->directory, rf->name, rf->extension);
	if ((fp = fopen(fn, "w")) == NULL) {
		logreturn(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in installfile(), fopen() error for %s", fn);
		free(fn);
		res = ERROR;
		return &res;
	}
	res = fwrite(rf->text, sizeof(char), rf->size, fp);
	if (res < rf->size) {
		if (res == 0)
			res = errno;
		else
			res = ERROR;
		logreturn(PIKTLOG, ERR, __FILE__, __LINE__, res == ERROR ? 0 : 1, "in installfile(), fwrite() error for %s", fn);
		if (fclose(fp)) {
			logreturn(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in installfile(), fclose() error for %s", fn);
			free(fn);
			res = ERROR;
			return &res;
		}
		return &res;
	}
	res = fclose(fp);
	if (res != 0) {
		logreturn(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in installfile(), fclose() error for %s", fn);
		free(fn);
		res = ERROR;
		return &res;
	}
	if ((res = chmod(fn, (mode_t)(rf->mode))) != 0) {
		logreturn(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in installfile(), chmod() error for %s", fn);
		free(fn);
		res = errno;
		return &res;
	}
	free(fn);
	return &res;
}

/* delete a file
 */
int *
DELETEFILE_1_SVC (rfil *rf, struct svc_req *rqstp)
{
	static char *fn;
	struct stat fbuf;
	static int res;

	if ((res = chkrequest(rqstp, "deletefile")) != 0)
		return &res;

	fn = (char *) malloc((strlen(rf->directory)+strlen(rf->name)+strlen(rf->extension)+2) * sizeof(char));
	sprintf(fn, "%s/%s%s", rf->directory, rf->name, rf->extension);
	if (stat(fn, &fbuf) == 0) {
		res = remove(fn);	/* remove returns 0 on success */
		if (res == ERROR) {
			logreturn(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in deletefile(), remove() error for %s", fn);
			res = errno;
		}
	} else	/* file not found, so signify success in removing it */
		res = 0;
	free(fn);
	return &res;
}

/* for both enablement and disablement, for the given alert, remove its
   line from the piktd.conf file (if present); for enablement, add a new
   line to that file for the given alert
 */
int *
EDITCONF_1_SVC (anam *an, struct svc_req *rqstp)
{
	FILE *pcfp, *tpcfp;
	char line[MAXLINE+1];
	char target[MAXLINE+1];
	static int res;

	if ((res = chkrequest(rqstp, "editconf")) != 0)
		return &res;

	if ((pcfp = fopen(piktdconf, "r")) == NULL) {
		logreturn(PIKTLOG, WARNING, __FILE__, __LINE__, TRUE, "in editconf(), fopen() error for %s", piktdconf);
		/* this might be a brand new installation, hence no prior
		   piktd.conf file
		 */
		/*
		res = ERROR;
		return &res;
		 */
	}
	if ((tpcfp = fopen(piktdconftmp, "w")) == NULL) {
		logreturn(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in editconf(), fopen() error for %s", piktdconftmp);
		res = ERROR;
		return &res;
	}
	if (pcfp) {
		sprintf(target, "+A %s", an->name);
		while (fgets(line, MAXLINE+1, pcfp) != NULL) {
			if (strstr(line, target) && line[0] != '#')
				continue;
			fputs(line, tpcfp);
		}
	}
	if (an->state == ADD)
		fprintf(tpcfp, "%s\n", an->aux);
	if (fclose(tpcfp)) {
		logreturn(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in editconf(), fclose() error for %s", piktdconftmp);
		res = ERROR;
		return &res;
	}
	if (pcfp) {
		if (fclose(pcfp)) {
			logreturn(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in editconf(), fclose() error for %s", piktdconf);
			res = ERROR;
			return &res;
		}
		if ((res = remove(piktdconf)) == ERROR) {	/* remove returns 0 on success */
			logreturn(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in editconf(), remove() error for %s", piktdconf);
			return &res;
		}
	}
	if ((res = rename(piktdconftmp, piktdconf)) == ERROR) {	/* 0 success, -1 failure */
		logreturn(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in editconf(), rename() error for %s and %s", piktdconftmp, piktdconf);
		return &res;
	}
	chmod(piktdconf, (mode_t)(S_IRUSR|S_IWUSR|S_IRGRP));
	res = 0;
	return &res;
}

/* kill the piktd daemon
 */
int *
KILLDAEMON_1_SVC (void *junk, struct svc_req *rqstp)
{
	pid_t oldpid, newpid;
	struct stat fbuf;
	char line[MAXLINE+1];
	FILE *fc;
	static int res;

	if ((res = chkrequest(rqstp, "killdaemon")) != 0)
		return &res;

	if ((oldpid = daemonpid()) == ERROR) {
		logreturn(PIKTLOG, ERROR, __FILE__, __LINE__, FALSE, "in killdaemon(), unable to kill piktd: multiple piktd's running");
		res = ERROR;
		return &res;
	} else if (oldpid == NONE) {
		res = NONE;
		return &res;
	}
	if (kill(oldpid, SIGKILL) != 0) {
		logreturn(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in killdaemon(), failure of kill command");
		res = ERROR;
		return &res;
	}
	if ((newpid = daemonpid()) != NONE) {
		logreturn(PIKTLOG, ERROR, __FILE__, __LINE__, FALSE, "in killdaemon(), unable to kill piktd (pid %d)", newpid);
		res = ERROR;
		return &res;
	} else
		logreturn(PIKTLOG, INFO, __FILE__, __LINE__, FALSE, "piktd (pid %d) killed", oldpid);
	/* killing piktd, and pikt runs, may yield a core file */
	if (lstat("/core", &fbuf) == 0) {
		if ((fc = popen(FILECMD " /core", "r")) == NULL) {
			logreturn(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in killdaemon(), popen() error");
		} else {
			fgets(line, MAXLINE+1, fc);
			pclose(fc);
			logreturn(PIKTLOG, WARNING, __FILE__, __LINE__, FALSE, "in killdaemon(), found /core file: %s", line);
		}
		if (remove("/core") == 0)	/* remove returns 0 on success */
			logreturn(PIKTLOG, WARNING, __FILE__, __LINE__, FALSE, "rm'ed /core file");
	}
	/* remove the piktd lock file */
	if (lstat(piktdlock, &fbuf) == 0) {
		if (remove(piktdlock) != 0)	/* remove returns 0 on success */
			logreturn(PIKTLOG, ERROR, __FILE__, __LINE__, TRUE, "in killdaemon(), unable to remove the %s file", piktdlock);
	}
	/* remove the piktd alerts file */
	if (lstat(piktdalt, &fbuf) == 0) {
		if (remove(piktdalt) != 0)	/* remove returns 0 on success */
			logreturn(PIKTLOG, ERROR, __FILE__, __LINE__, TRUE, "in killdaemon(), unable to remove the %s file", piktdalt);
	}
	res = 0;
	return &res;
}

/* restart the piktd daemon
 */
int *
RESTARTDAEMON_1_SVC (void *junk, struct svc_req *rqstp)
{
	pid_t oldpid, newpid;
	pid_t pid;
	struct stat fbuf;
	static int res;

	if ((res = chkrequest(rqstp, "restartdaemon")) != 0)
		return &res;

	if ((oldpid = daemonpid()) == NONE) {
		/* if no piktd's running, remove any orphaned piktd lock
		   files
		 */
		if (lstat(piktdlock, &fbuf) == 0) {
			if (remove(piktdlock) != 0) {	/* remove returns 0 on success */
				logreturn(PIKTLOG, ERROR, __FILE__, __LINE__, TRUE, "in restartdaemon(), unable to restart piktd: unable to remove %s file", piktdlock);
				res = ERROR;
				return &res;
			}
		}
	} else if (oldpid == ERROR) { /* we have multiple piktd's running */
		logreturn(PIKTLOG, ERROR, __FILE__, __LINE__, TRUE, "in restartdaemon(), unable to restart piktd: multiple piktd's running");
		res = ERROR;
		return &res;
	} else { /* we have a running piktd */
		logreturn(PIKTLOG, ERROR, __FILE__, __LINE__, TRUE, "in restartdaemon(), unable to restart piktd: another piktd (pid %d) running", oldpid);
		res = ERROR;
		return &res;
	}
	switch (pid = fork()) {
		case -1:	/* error */
			logexit(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in restartdaemon(), fork() error");
			break;
		case 0:		/* child */
			execl(SHELL, "sh", "-c", piktd, (char *) NULL);
			_exit(127);
			break;
		default:	/* parent */
			#if defined(TIME_TO_SLEEP)
			while (!waitpid(pid, NULL, WNOHANG))
				sleep(TIME_TO_SLEEP);
			#else
			waitpid(pid, NULL, 0);
			#endif
			break;
	}
	newpid = daemonpid();
	if ((newpid == ERROR) || (newpid == oldpid)) {
		logreturn(PIKTLOG, ERR, __FILE__, __LINE__, FALSE, "in restartdaemon(), failed to restart piktd");
		res = ERROR;
		return &res;
	}
	logreturn(PIKTLOG, INFO, __FILE__, __LINE__, FALSE, "piktd restart");
	res = 0;
	return &res;
}

/* report the status of the given alert; an alert is inactive if: there is
   no .alt file, there is no line for that alert in piktd.conf, there is
   no active piktd, and/or the alert is absent from the piktd alerts file
 */
int *
SHOWSTATUS_1_SVC (anam *an, struct svc_req *rqstp)
{
	FILE *pcfp, *rfp;
	static char *fn;
	char line[MAXLINE+1];
	char target[MAXLINE+1];
	struct stat fbuf;
	static int res;
	char *cp;

	if ((res = chkrequest(rqstp, "showstatus")) != 0)
		return &res;

	/* is there an .alt file? */
	fn = (char *) malloc((strlen(altdir)+strlen(an->name)+6) * sizeof(char));
	sprintf(fn, "%s/%s.alt", altdir, an->name);
	if ((res = stat(fn, &fbuf)) == ERROR) {	/* 0 success, -1 failure */
		logreturn(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in showstatus(), stat() error for %s", fn);
		free(fn);
		return &res;
	}

	/* found in piktd.conf? */
	if ((pcfp = fopen(piktdconf, "r")) == NULL) {
		logreturn(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in showstatus(), fopen() error for %s", piktdconf);
		res = ERROR;
		return &res;
	}
	sprintf(target, "+A %s", an->name);
	res = ERROR;
	while (fgets(line, MAXLINE+1, pcfp) != NULL) {
		if (strstr(line, target) && line[0] != '#') {
			res = 0;
			break;
		}
	}
	if (fclose(pcfp)) {
		logreturn(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in showstatus(), fclose() error for %s", piktdconf);
		res = ERROR;
		return &res;
	}
	if (res == ERROR) {
		logreturn(PIKTLOG, ERR, __FILE__, __LINE__, FALSE, "in showstatus(), %s entry not found in %s", an->name, piktdconf);
		return &res;
	}

	/* active piktd? */
	if ((res = daemonpid()) == ERROR) {
		logreturn(PIKTLOG, ERR, __FILE__, __LINE__, FALSE, "in showstatus(), active piktd not found");
		return &res;
	}

	/* registered in active pikd? */
	if ((rfp = fopen(piktdalt, "r")) == NULL) {
		logreturn(PIKTLOG, ERR, __FILE__, __LINE__, TRUE, "in showstatus(), fopen() error for %s", piktdalt);
		res = ERROR;
		return &res;
	}
	res = ERROR;
	while (fgets(line, MAXLINE+1, rfp) != NULL) {
		if ((cp = strchr(line, '\n')) != NULL)
			*cp = '\0';
		if (!strcmp(line, an->name)) {
			res = 0;
			break;
		}
	}
	if (res == ERROR) {
		logreturn(PIKTLOG, ERR, __FILE__, __LINE__, FALSE, "in showstatus(), %s not found in %s", an->name, piktdalt);
		return &res;
	}

	return &res;
}

/* return the piktd pid, or NONE if no daemon is running, or ERROR if
   multiple piktd's are running; when determining this, ignore piktd's
   spawned for launching pikt runs

   we used to invoke ioctl() on each /proc/<pid> to get the process names,
   but we saw several cases where the resultant process name for a given
   pid didn't match the process name for that pid as shown in the ps display;
   the technique of reading the ps display, although perhaps less elegant,
   gives us the consistent results we desire
 */
pid_t
daemonpid (void)
{
	pid_t pspid;
	long pid, ppid;
	#if PSPARSE == 1
	char psp[MAXLINE+1];
	#endif
	char *psproc;
	char line[MAXLINE+1];
	FILE *ps;
	char *cp;

	pspid = 0;
	if ((ps = popen(PSCMD " " PSOPTS, "r")) == NULL) {
		logreturn(PIKTLOG, ERR, __FILE__, __LINE__, FALSE, "in daemonpid(), popen() error");
		return ERROR;
	}
	fgets(line, MAXLINE+1, ps);	/* throw away header */
	while (fgets(line, MAXLINE+1, ps)) {
		#if PSPARSE == 1
		if (sscanf(line, "%ld %ld %s", &pid, &ppid, psp) != 3) {
			logreturn(PIKTLOG, WARNING, __FILE__, __LINE__, FALSE, "in daemonpid(), error reading pid/ppid/psproc for line %s", line);
			continue;
		}
		/* strip away path, if any, leaving only basename */
		if ((cp = strrchr(psp, '/')))
			psproc = cp+1;
		else
			psproc = psp;
		#else
		# if PSPARSE == 2
		if (sscanf(line, "%*s %*s %ld %ld %*s %*s %*s %*s %*s %*s %*s %*s %*s", &pid, &ppid) != 2) {
			logreturn(PIKTLOG, WARNING, __FILE__, __LINE__, FALSE, "in daemonpid(), error reading pid/ppid for line %s", line);
			continue;
		}
		if ((psproc = strrchr(line, ':')))
			psproc += 4;
		else {
			logreturn(PIKTLOG, WARNING, __FILE__, __LINE__, FALSE, "in daemonpid(), error reading psproc for line %s", line);
			continue;
		}
		# else
		#  include "Bletch: How do we parse `ps' output?"
		# endif
		#endif
		/* for some strange reason, linux ps appends a space to the
		   end of each ps line
		 */
		if ((!strcmp(psproc, "piktd") || !strncmp(psproc, "piktd\n", 6) || !strncmp(psproc, "piktd ", 6)) && (ppid == 1)) {
			if (pspid == 0)
				pspid = (pid_t) pid;
			else {
				logreturn(PIKTLOG, ERR, __FILE__, __LINE__, FALSE, "failed to determine unique piktd pid: multiple piktd's running");
				pclose(ps);
				return ERROR;
			}
		}
	}
	pclose(ps);
	if (pspid == 0) {
		logreturn(PIKTLOG, WARNING, __FILE__, __LINE__, FALSE, "failed to determine piktd pid: piktd not running");
		return NONE;
	} else
		return pspid;
}
