/*
   common.h -- PIKT common header file for pikt & piktd

   $Header: /home/cvs/ucgsb/pikt/src/piktd/common.h,v 1.14 1999/04/14 14:22:56 harlan Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <time.h>
#include <sys/types.h>
#include <sys/time.h>
#ifndef FILE
#include <stdio.h>
#endif
#include <rxposix.h>
#include <unistd.h>

#define MAXLINE		512	/* max length of input data line, or max
				   line length in various other contexts
				 */
#define MAXCMD		MAXLINE	/* max length of scancmd, also for exec'ed
				   commands; a cmd longer than this will
				   dump core; we don't error check this, but
				   make this generously long
				 */
#define MAXKEY		MAXLINE	/* max length of key value; we set this to
				   MAXLINE, which is way generous
				 */

#define MAXBUF		65536	/* max length of function buffer string */
#define MAXPATH		256	/* max length of file path */
#define MAXIDENT	256	/* max length of identifier */

#define MAXIFS		64	/* max number of nested if's */
#define MAXWHILES	MAXIFS
#define MAXREPEATS	MAXIFS
#define MAXFORS		MAXIFS
#define MAXSWITCHS	MAXIFS
#define MAXLOOPS	3*MAXIFS

#define MAXVAP		64	/* MAXVAP, the maximum level of array
				   nesting, is an arbitrarily large
				   figure for its context
				 */
#define MAXFNN		64	/* MAXFNN, the maximum level of function
				   nesting, is an arbitrarily large
				   figure for its context
				 */

#define MAXVARS		255	/* max number of user-defined var names */

#define MAXSUBARY	256	/* max num sub arrays */
#define MAXSUBNDX	256	/* max sub array index */
#define MAXARYNDX	65535	/* max array index, MAXSUBARY*MAXSUBNDX - 1 */

#define MAXDBLSTR	320	/* max length of double string */
#define MAXFLTSTR	48	/* max length of float string */
#define MAXINTSTR	16	/* max length of int string */

#define EXECWAITWAIT	1	/* arbitarily long pause (in seconds)
				   between checks for existence of exec
				   lock file when doing an exec wait
				 */

#define ERR		-1	/* also used for log type */
#define NIL		-2	/* for value unknown */
#define DENIED		-3	/* denied piktc service request */

#ifndef FALSE
#define FALSE		0
#endif
#ifndef TRUE
#define TRUE		1
#endif

#define VLNOW		-1	/* $ and # values */
#define VLHST		-2	/* % history values */
#define VLPRC		-3	/* @ preceding values */
#define VLPRS		-4	/* regexp and #parse() values */
#define INLINNOW	-5	/* $inlin */
#define INLINPRC	-6	/* @inlin */
#define RDLINNOW	-7	/* $rdlin */
#define RDLINPRC	-8	/* @rdlin */

#define SQTPROXY	-127	/* single-quote stand-in */
#define DQTPROXY	-128	/* double-quote stand-in */

/* status */
#define INACTIVE	0
#define ACTIVE		1

/* input type */
#define NOINPUT		0
#define LOGINPUT	1	/* must be 1; ands/ors with val types below */
#define FILEINPUT	2
#define PROCINPUT	3

/* val type */
#define FIXEDVAL	2	/* var referenced in init dat statement */
#define FREEVAL		4	/* var referenced in rule section */
#define BEGINVAL	8	/* var referenced in begin section */
#define ENDVAL		16	/* var referenced in end section */

/* dat type */
#define STRING		0
#define NUMBER		1
#define FILEHANDLE	2
#define PROCHANDLE	3

/* expr type */
#define RHS		0	/* right-hand side */
#define	LHS		1	/* left-hand side */

/* alarm level -- correspond to syslog levels */
#define EMERG		0
#define ALERT		1
#define CRIT		2
/* #define ERR		3	   see above              */
#define WARNING		4	/* also used for log type */
#define NOTICE		5
#define INFO		6	/* also used for log type */
#define DEBUG		7	/* also used for log type */

/* statement type */
#define DO		1
#define IF		2
#define ELIF		3
#define ELSE		4
#define ENDIF		5
#define WHILE		6
#define ENDWHILE	7
#define WHILEAUX	107
#define REPEAT		8
#define UNTIL		9
#define REPEATAUX	109
#define FOR		10
#define ENDFOR		11
#define FORAUX		111
#define BREAK		12
#define CONTINUE	13
#define SWITCH		14
#define CASE		15
#define DEFAULT		16
#define BREAKSWITCH	17
#define ENDSWITCH	18

#define AGAIN		19
#define REDO		20
#define NEXT		21
#define SKIP		22
#define LAST		23
#define PAUSE		24
#define QUIT		25
#define DIE		26

#define OUTPUT		27
#define MAIL		28
#define PRINT		29
#define LOCLOG		30
#define PIKTLOG		31
#define SYSLOG		32

#define SET		33
#define EXEC		34
#define EXECWAIT	35

/* rul type */
#define BEG		1
#define RUL		2
#define END		3

/* used in scan functions */
#define BEGINDAT	1
#define INDAT		2
#define ENDDAT		3
#define OUTDAT		4

/* pos type */
#define FIELDS		1
#define COLUMNS		2
#define REGEXPR		3

/* filestat type */
#define EXIST		0
#define ZEROLEN		1
#define REGFILE		2
#define DIRECTORY	3
#define CHARSPEC	4
#define BLKSPEC		5
#define PIPE		6
#define SYMLINK		7
#define SOCKET		8

/* trim type */
#define NONE		0
#define LEFT		1
#define RIGHT		2
#define BOTH		3

/* sys type */
#define SYSTEM		0
#define COMMAND		1

/* var type */
#define SCALAR		0
#define ARRAY		1

/* procid type */
#define PROCID		0
#define PPROCID		1

/* some extremely useful macros follow, without which the code would be
   much more hard to read

   my entries in the obfuscated code contest:
 */

#define VALGLB		alarmcurr->values->next->next
#define VALHST		alarmcurr->values->next->next->next

#define ISSCALAR(ndx)	alarmcurr->vartype[ndx] == SCALAR
#define ISARRAY(ndx)	alarmcurr->vartype[ndx] == ARRAY

#define SA(a)	(int)((a-1)/MAXSUBNDX)
#define SN(a)	(a-1)%MAXSUBNDX

/* reference the fldnum */
#define FLDNUM(vndx,andx)	(andx ? (alarmcurr->vars[vndx].a->subary[SA(andx)] ? alarmcurr->vars[vndx].a->subary[SA(andx)]->fldnum[SN(andx)] : 0) : alarmcurr->vars[vndx].s->fldnum[andx])

/* reference the begincol */
#define BEGINCOL(vndx,andx)	(andx ? (alarmcurr->vars[vndx].a->subary[SA(andx)] ? alarmcurr->vars[vndx].a->subary[SA(andx)]->begincol[SN(andx)] : 0) : alarmcurr->vars[vndx].s->begincol[andx])

/* reference the endcol */
#define ENDCOL(vndx,andx)	(andx ? (alarmcurr->vars[vndx].a->subary[SA(andx)] ? alarmcurr->vars[vndx].a->subary[SA(andx)]->endcol[SN(andx)] : 0) : alarmcurr->vars[vndx].s->endcol[andx])

/* reference whether trim */
#define TRIM(vndx,andx)		(andx ? (alarmcurr->vars[vndx].a->subary[SA(andx)] ? alarmcurr->vars[vndx].a->subary[SA(andx)]->trim[SN(andx)] : NIL) : alarmcurr->vars[vndx].s->trim[andx])

/* reference a scalar value */
#define VALSCL(vndx,andx)		vals[vndx].s->val[andx]

/* reference a subarray */
#define SUBARY(vndx,andx)		vals[vndx].a->subary[SA(andx)]

/* reference an array value */
#define VALARY(vndx,andx)		vals[vndx].a->subary[SA(andx)]->val[SN(andx)]

/* set a var spec */
#define SETSPC(spc,vndx,andx,vl)	if (andx) {\
						if (! alarmcurr->vars[vndx].a->subary[SA(andx)])\
							initspcsubary(&alarmcurr->vars[vndx].a->subary[SA(andx)]);\
						alarmcurr->vars[vndx].a->subary[SA(andx)]->spc[SN(andx)] = vl;\
					} else\
						alarmcurr->vars[vndx].s->spc[0] = vl

/* set a value */
#define SETVAL(valrec,vndx,andx,vl)	if (andx) {\
						if (! valrec->vals[vndx].a->subary[SA(andx)]) {\
							initchrsubary(&valrec->vals[vndx].a->subary[SA(andx)]);\
						}\
						valrec->vals[vndx].a->subary[SA(andx)]->val[SN(andx)] = vl;\
					} else\
						valrec->vals[vndx].s->val[0] = vl
					
#define ALMNAM		alarmcurr->name
#define STATUS		alarmcurr->status
#define LEVEL		alarmcurr->level
#define TASK		alarmcurr->task
#define INPUTTYP	alarmcurr->input.type
#define INPUTSRC	alarmcurr->input.source
#define INPUTLOGNAME	alarmcurr->input.logname
#define INPUTLOGINO	alarmcurr->input.logino
#define INPUTLOGSIZE	alarmcurr->input.logsize
#define INPUTLOGPOSEND	alarmcurr->input.logposend
#define SEPARATORS	alarmcurr->separators
#define FILTER		alarmcurr->filter
#define DATCNT		alarmcurr->datcnt
#define REGEXP		alarmcurr->regexp
#define MAXDATNDX	alarmcurr->maxdatndx
#define POSTYP		alarmcurr->postype
#define VALUES		alarmcurr->values
#define SCVARS		alarmcurr->scvars
#define KEYS		alarmcurr->keys
#define HSTTYP		alarmcurr->hsttyp
#define HSTERR		alarmcurr->hsterr

#define VARNAM(ndx)	alarmcurr->varname[ndx]
#define VALTYP(ndx)	alarmcurr->valtype[ndx]
#define VARTYP(ndx)	alarmcurr->vartype[ndx]
#define DATTYP(ndx)	alarmcurr->dattype[ndx]
#define MAXARNDX(ndx)	alarmcurr->maxarndx[ndx]
#define OLDHST(ndx)	alarmcurr->oldhst[ndx]
#define NEWHST(ndx)	alarmcurr->newhst[ndx]

#define FLHNDL(ndx)	alarmcurr->vars[ndx].f

#define SHELL		"/bin/sh"	/* for purposes of execl in
					   scanopen() and dostm()
					 */

#define foreach(a,b)	for (a=b; a; a=a->next)
#define maxval(a,b)	a >= b ? a : b
#define minval(a,b)	a <= b ? a : b
#define odd(a)		(a % 2) ? TRUE : FALSE
#define even(a)		(a % 2) ? FALSE : TRUE

#define FREE(a)		if (a) free(a)

/* for queued messages to be e-mailed or printed */
typedef struct _msg {
	char *line;
	struct _msg *next;
} msg;

/* for history value lookup keys */
typedef struct _ky {
	int vrndx;	/* var index */
	int arndx;	/* array index */
	struct _ky *next;
} ky;

/* for stat functions */
typedef struct _st {
	double num;		/* a value */
	int freq;		/* its frequency of occurence */
	struct _st *next;
} st;

/* for recording input info */
typedef struct _inp {
	int type;
	char *source;			/* input-generating command string */
	char *logname;			/* input log's name */
	unsigned long logino;		/* input log's inode */
	unsigned long logsize;		/* input log's size in bytes */
	unsigned long logposend;	/* position of last read byte in
					   input log
					 */
} inp;

/* for storing alarm script tokens */
typedef struct _tok {
	char *token;
	struct _tok *next;
} tok;

/* for recording statement info */
typedef struct _stm {
	int type;		/* DO, IF, ELIF, ... */
	int status;		/* ACTIVE/INACTIVE */
	tok *rhsexpr;		/* expr on right side of = in set statement */
	int vrndx;		/* index of var on left side of set
				   statement
				 */
	tok *lhsexpr;		/* array index expr on left side of set
				   statement
				 */
	int arndx;		/* array index on left side of set statement */
	char *logfile;		/* logfile tied to output log statement */
	struct _stm *tnext;	/* next statement if current statement
				   is true
				 */
	struct _stm *fnext;	/* next statement if current statement
				   is false
				 */
	struct _stm *aux;	/* auxiliary/dummy statement */
	struct _stm *next;	/* next linked statement, defaults to tnext */
} stm;

/* list of rules */
typedef struct _rul {
	stm *statements;
	struct _rul *next;
} rul;

/* val scalar */
typedef struct {
	char *val[1];
} chrscl;

/* val subarray */
typedef struct {
	char *val[MAXSUBNDX];
} chrsubary;

/* val array */
typedef struct {
	chrsubary *subary[MAXSUBARY];
} chrary;

/* list of vals */
typedef struct _val {
	char *key;		/* history lookup key val */
	union {
		chrscl *s;	/* used if scalar */
		chrary *a;	/* used if array */
	} vals[MAXVARS+1];
	struct _val *prev;
	struct _val *next;
} val;

/* dat info scalar */
typedef struct {
	int fldnum[1];		/* the field number */
	int begincol[1];	/* its begin column, if columnar */
	int endcol[1];		/* its end column, if columnar */
	int trim[1];		/* trim leading/trailing whitespace? */
} spcscl;

/* dat info subarray */
typedef struct {
	int fldnum[MAXSUBNDX];
	int begincol[MAXSUBNDX];
	int endcol[MAXSUBNDX];
	int trim[MAXSUBNDX];
} spcsubary;

/* dat info array */
typedef struct {
	spcsubary *subary[MAXSUBARY];
} spcary;

/* alarm info */
typedef struct _alrm {
	char *name;			/* alarm name */
	int status;			/* ACTIVE/INACTIVE */
	int level;			/* EMERG/ALERT/CRIT/... */
	char *task;			/* alarm description */
	inp input;			/* input info */
	char *separators;		/* dat field separators, if any */
	char *filter;			/* input filter, if any */
	char *regexp;			/* regexp, if dat "<regexp>" */
	int datcnt;			/* count of dat items */
	int maxdatndx;			/* highest dat index */
	int postype;			/* FIELDS/COLUMNS/REGEXPR */
	char *varname[MAXVARS+1];	/* var names */
	int valtype[MAXVARS+1];		/* FIXEDVAL/FREEVAL/BEGINVAL/ENDVAL */
	int vartype[MAXVARS+1];		/* SCALAR/ARRAY */
	int dattype[MAXVARS+1];		/* STRING/NUMBER/FILEHANDLE/PROCHANDLE */
	int maxarndx[MAXVARS+1];	/* highest array index recorded */
	int oldhst[MAXVARS+1];		/* found in old history file */
	int newhst[MAXVARS+1];		/* record in new history file */
	union {
		spcscl *s;		/* scalar */
		spcary *a;		/* array */
		FILE   *f;		/* filehandle */
	} vars[MAXVARS+1];
	val *values;			/* links to vals list */
	ky *scvars;			/* scancolumn vars list */
	ky *keys;			/* history lookup keys */
	rul *begin;			/* begin rule (one only) */
	rul *rules;			/* list of rules */
	rul *end;			/* end rule (one only) */
	int hsttyp;			/* FIXEDVAL|FREEVAL|BEGINVAL|ENDVAL */			
	int hsterr;			/* error in history processing */
	struct _alrm *next;
} alrm;

/* alert info */
typedef struct _alt {
	unsigned long himin;	/* 32-59 */
	unsigned long lomin;	/*  0-31 */
	unsigned long hr;	/*  0-23 */
	unsigned long dom;	/*  1-31 */
	unsigned long mon;	/*  1-12 */
	unsigned long dow;	/*  0- 6 */
	char *piktcmd;
	struct _alt *next;
} alt;

/* dir/file paths */
extern char *home,		/* PIKT home directory */
            *bindir,		/* binaries directory */
            *etcdir,		/* admin directory */
            *libdir,		/* lib directory */
            *vardir,		/* var directory */
            *altdir,		/* alerts directory */
            *prgdir,		/* programs directory */
            *objdir,		/* objects directory */
            *cfgdir,		/* configs directory */
            *stgdir,		/* staging directory */
            *hstdir,		/* histories directory */
            *logdir,		/* log directory */
            *PIKTconf,		/* PIKT conf file */
            *pikt,		/* pikt program */
            *piktd,		/* piktd program */
            *piktdconf,		/* piktd conf file */
            *piktdconftmp,	/* piktd conf temp file */
            *piktdlock,		/* piktd lock file */
            *piktdlog,		/* piktd log file */
            *piktdalt,		/* piktd alerts file */
            *piktclock,		/* piktc lock file */
            *piktclog,		/* piktc log file */
            *piktc_svclock,	/* piktc_svc lock file */
            *piktc_svclog;	/* piktc_svc log file */

/* command-line options */
extern int ischeck,
           isdebug,
           isdoublecheck,
           ishelp,
           istest;

extern int interactive;

extern char **argvcurr;	/* current command-line argument */
extern char **argvlast; /* last command-line argument */

extern unsigned stroffset; /* offset from beginning of current string */

extern const char pikt_version[];
extern const char piktd_version[];

extern int uid;			/* piktc/piktc_svc service request uid */
extern char *master;		/* hostname of master control system */
extern char *domain;		/* local network domain, or "domain" */
extern char *accessmode;	/* "local" or "global" */

extern char hostname[];			/* name of alert host */
extern char domainname[];		/* domainname of alert host */

extern int errno;

extern char *altcurr;			/* current alert name */

extern char *altlockfile;		/* name of lock file for this alert */
extern char *execlockfile;		/* name of exec lock file for this alert */
extern char *piktlog;			/* name of piktlog for this alert */
/* extern char syslog[]; */		/* path of syslog file */

extern const char *leveltypes[];	/* EMERGENCY/ALERT/CRITICAL/... */

extern char inlin[][];			/* input lines */
extern char *inlinnow, *inlinprc;	/* ptrs to current/preceding inlin */
extern int innum;			/* number of current input line */
extern int inlen;			/* length of current input line */
extern char rdlin[][];			/* output lines from #read() */
extern char *rdlinnow, *rdlinprc;	/* ptrs to current/preceding rdlin */
extern int skip;			/* number of input lines to skip */
extern int eoi;				/* is end-of-input? */

extern int numcnt;			/* count of stats; used by stat fns */
extern double numsum;			/* sum of stats; used by stat fns */
extern double nummax;			/* max of stats; used by stat fns */
extern double nummin;			/* min of stats; used by stat fns */
extern st *stats;			/* list of stats; used by stat fns */

extern time_t nextmin;

/* for timing */
extern char *tcp;
extern unsigned long cb;
extern int clo, chi, cmin, cmax;	/* timing bit parameters */

extern unsigned long himincurr;		/* 32 bits, for seconds 0-31 */
extern unsigned long lomincurr;		/* 32 bits, for seconds 32-59 */
extern unsigned long hrcurr;		/* current hour */
extern unsigned long domcurr;		/* current day of month */
extern unsigned long moncurr;		/* current month */
extern unsigned long dowcurr;		/* current day of week */

extern int prescan;		/* is this script scan preliminary or not? */

extern int lineno;		/* line number in alert script file */

extern int dohistory;		/* do history processing? */

extern int isprc;		/* have preceding vals? */

extern int exprtyp;		/* LHS/RHS */

extern int doeval;		/* evaluate this expression? */
extern int exprres;		/* TRUE/FALSE/ERR */
extern double exprval;		/* eval'ed expression as a number */
extern char *exprstr;		/* eval'ed expression as a string */

extern int dpls;		/* decimal places */

extern regmatch_t regmatch[MAXARYNDX+1];	/* matched ()'ed expressions */
extern int matches;			/* number of matched ()'ed expressions */

extern char *arytyp[];		/* ARYSTR/ARYNUM */

extern char fnnumnames[];	/* string of all number function names */
extern char fnstrnames[];	/* string of all string function names */

extern char nil[];		/* unknown val */
extern char err[];		/* error val */
extern const char nilstr[];	/* unknown val */
extern const char errstr[];	/* error val */
extern int iserrdef;		/* is #error() or #defined() function? */

extern const char dquote[];	/* double quote */
extern const char squote[];	/* single quote */
extern const char space[];

extern char *mailcmd, *lpcmd;		/* mail/print commands */
extern msg *mailmsg, *printmsg;		/* queued mail/print messages */

extern FILE *hstout;			/* history output */
extern char *hsttmpfile;		/* temp (new) history file */

extern alrm *alarms;			/* list of alarms */

extern int newmailalarm, newmailalert, newprintalarm, newprintalert;

extern alrm *alarmcurr;		/* current alarm */
extern val *valcurr;		/* current val */
extern ky *scvarcurr;		/* current scancolumn var */
extern ky *keycurr;		/* current key */
extern rul *rulcurr;		/* current rule */
extern tok *tokcurr;		/* current token */
extern stm *stmcurr;		/* current statement */

extern val *valnow;		/* val in current time context */
extern val *valprc;		/* val for preceding input line */
extern val *valtmp;		/* temporary val */
extern val *valglb;		/* "global" val */
extern val *valhst;		/* val when alarm last run */

extern char *valscn[MAXARYNDX+1];	/* scanned values */

extern chrary *vap[MAXVAP];	/* pointer to val array */
extern int vapndx;		/* array index */

/* init functions */
extern void init(int, char **, void (*)(void));

/* command-line functions */
extern void pikt_usage(void);
extern void piktd_usage(void);

/* lexer and parser functions */
extern int cmdparse(void);
extern int altparse(void);
extern int exprparse(void);
extern int hstparse(void);
extern int timingparse(void);
extern int cmdlex(void);
extern int altlex(void);
extern int exprlex(void);
extern int hstlex(void);
extern int timinglex(void);
extern void cmderror(const char *);
extern void alterror(const char *);
extern void exprerror(const char *);
extern void hsterror(const char *);
extern void timingerror(const char *);

/* alt file functions */
extern void regalm(char *);
extern void regvar(char *, int, int, int, int, char *);
extern void regscvar(int, int);
extern void regkey(char *, int, int, int);
extern void regrul(int);
extern void regtok(char *, int);
extern void duptok(tok *, tok **);
extern void regstm(int);
extern void reghst(int);
extern int varndx(char *);
extern int vardattype(char *);
extern void initchrsubary(chrsubary **);
extern void initspcsubary(spcsubary **);

/* expr functions */
extern time_t now(void);
extern time_t today(void);
extern time_t datevalue(int, int, int);
extern time_t timevalue(int, int, int);
extern int second(time_t);
extern int minute(time_t);
extern int hour(time_t);
extern int day(time_t);
extern int month(time_t);
extern int year(time_t);
extern int weekday(time_t);
extern int yearday(time_t);
extern int yearweek(time_t);
extern char *ampm(time_t);
extern int monthnumber(char *);
extern const char *monthname(int);
extern int daynumber(char *);
extern const char *dayname(int);
extern int daysbetween(time_t, time_t);
extern int age(char *, char *, char *);
extern int filestat(char *, int);

extern int ceiling(double);
extern int integer(double);
extern int round(double);
extern int trunc(double);
extern int isint(double);
extern void regstat(double);
extern void freestats(void);
extern double median(void);
extern double mode(void);

extern char *my_char(int);
extern char *chop(char *, int);
extern int code(char *);
extern char *fixed(double, int);
extern int search(char *, char *, int);
extern int rsearch(char *, char *);
extern char *reverse(char *);
extern char *lower(char *);
extern char *proper(char *);
extern char *repeat(char *, int);
extern char *replace(char *, char *, int, int);
extern char *substitute(char *, char *, char *, int);
extern char *substring(char *, int, size_t);
extern char *sys(char *, int);
extern char *trim(char *, int);
extern char *upper(char *);

extern char *my_basename(char *);
extern char *my_dirname(char *);

/* history functions */
extern val *lookupval(char *, val *);
extern alrm *lookupalarm(char *);
extern void verifylog(char *);
extern void verifykey(int, int);
extern void verifyvar(int, char *);
/* extern void verifyfld(int, int, int); */
extern void verifytyp(int, int);
extern void verifyval(int, int);
extern void verifyglb(int, int);
extern void regval(int, int, char *);
extern void makekeys(void);
extern void freevals(val **, int, int);
extern void freevalhst(val *);
extern void freevalhsts(val **);

/* timing functions */
extern void parsetiming(char *, unsigned long *, int, int, int, int);
extern void timingsync(void);
extern void timingsleep(void);

/* statement functions */
extern int dostm(int);
extern void sendmsg(msg *, char *);
extern void freemsg(msg **);

/* scan functions */
extern int scan(int, char *, int);
extern int scanfields (char *, char *, int);
extern FILE *scanopen(char *, pid_t *);
extern void scanclose(FILE *, pid_t);
extern char *stripesc(char *);
extern int dequote(char *);
extern int requote(char *);
extern void freevalscn(void);

/* regexp functions */
extern int regexp(char *, char *, int, int);

/* proc functions */
extern pid_t procid(char *, int);

/* lock functions */
extern int islock(char *);
extern int mklock(char *, char *, int, int);
extern int rmlock(char *, int, int);
extern int rmlocks(int, int);

/* logging functions */
extern void logreturn(int, int, const char *, int, int, const char *, ...);
extern void logexit(int, int, const char *, int, int, const char *, ...);

/* signal functions */
extern void sig_init(void);

/* debugging functions */
/* extern void dumpalarms(void); */
extern void dumpvalnows(val *);
extern void dumpvals(val *);
extern void dumpsps(stm *, char *);
/* extern void dumptime(char *, unsigned long); */
